# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests of the scenario configs."""

import collections

from absl.testing import absltest
from absl.testing import parameterized

from meltingpot import bot as bot_factory
from meltingpot.configs import bots
from meltingpot.configs import scenarios
from meltingpot.configs import substrates

SCENARIO_CONFIGS = scenarios.SCENARIO_CONFIGS
AVAILABLE_BOTS = bot_factory.BOTS
AVAILABLE_SUBSTRATES = frozenset(substrates.SUBSTRATES)


def _is_compatible(bot_name, substrate, role):
  if bot_name == bot_factory.NOOP_BOT_NAME:
    return True
  bot_config = bots.BOT_CONFIGS[bot_name]
  return substrate == bot_config.substrate and role in bot_config.roles


class ScenarioConfigTest(parameterized.TestCase):

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_description(self, scenario):
    self.assertNotEmpty(scenario.description)

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_tags(self, scenario):
    self.assertNotEmpty(scenario.tags)

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_valid_substrate(self, scenario):
    self.assertIn(scenario.substrate, AVAILABLE_SUBSTRATES)

  @parameterized.named_parameters(
      (name, name, scenario) for name, scenario in SCENARIO_CONFIGS.items())
  def test_name_starts_with_substrate_name(self, name, scenario):
    self.assertStartsWith(name, scenario.substrate)

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_focal_players(self, scenario):
    self.assertTrue(any(scenario.is_focal))

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_matching_sizes(self, scenario):
    self.assertLen(scenario.is_focal, len(scenario.roles))

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_valid_roles(self, scenario):
    valid_roles = substrates.get_config(scenario.substrate).valid_roles
    self.assertContainsSubset(scenario.roles, valid_roles)

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_has_valid_bots(self, scenario):
    scenario_bots = set().union(*scenario.bots_by_role.values())
    self.assertContainsSubset(scenario_bots, AVAILABLE_BOTS)

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_bots_compatible(self, scenario):
    for role, bot_names in scenario.bots_by_role.items():
      incompatible = {
          bot_name for bot_name in bot_names
          if not _is_compatible(bot_name, scenario.substrate, role)
      }
      with self.subTest(role):
        self.assertEmpty(
            incompatible,
            f'Substrate {scenario.substrate!r}, role {role!r} not supported '
            f'by: {incompatible!r}.')

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_no_missing_role_assigments(self, scenario):
    background_roles = set(role for n, role in enumerate(scenario.roles)
                           if not scenario.is_focal[n])
    supported_roles = {
        role for role, bots in scenario.bots_by_role.items() if bots}
    unsupported_roles = background_roles - supported_roles
    self.assertEmpty(unsupported_roles,
                     f'Background roles {unsupported_roles!r} have not been '
                     f'assigned bots.')

  @parameterized.named_parameters(SCENARIO_CONFIGS.items())
  def test_no_unused_role_assignments(self, scenario):
    background_roles = set(role for n, role in enumerate(scenario.roles)
                           if not scenario.is_focal[n])
    redundant_roles = set(scenario.bots_by_role) - background_roles
    self.assertEmpty(redundant_roles,
                     f'Bots assigned to {redundant_roles!r} are unused.')

  def test_no_duplicates(self):
    seen = collections.defaultdict(set)
    for name, config in SCENARIO_CONFIGS.items():
      seen[config].add(name)
    duplicates = {names for _, names in seen.items() if len(names) > 1}
    self.assertEmpty(duplicates, f'Duplicate configs found: {duplicates!r}.')

  def test_all_substrates_used_by_scenarios(self):
    used = {scenario.substrate for scenario in SCENARIO_CONFIGS.values()}
    unused = AVAILABLE_SUBSTRATES - used
    self.assertEmpty(unused, f'Substrates not used by any scenario: {unused!r}')

  def test_all_bots_used_by_scenarios(self):
    used = set()
    for scenario in SCENARIO_CONFIGS.values():
      used.update(*scenario.bots_by_role.values())
    unused = AVAILABLE_BOTS - used
    self.assertEmpty(unused, f'Bots not used by any scenario: {unused!r}')


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mocks of various Melting Pot classes for use in testing."""

from collections.abc import Mapping, Sequence
from typing import Optional, Type, TypeVar
from unittest import mock

import dm_env
import immutabledict
import numpy as np
import tree

import meltingpot
from meltingpot.utils.scenarios import scenario
from meltingpot.utils.substrates import specs as meltingpot_specs
from meltingpot.utils.substrates import substrate

SUBSTRATE_OBSERVATION_SPEC = immutabledict.immutabledict({
    # Observations present in all substrates. Sizes may vary.
    'RGB': meltingpot_specs.OBSERVATION['RGB'],
    'WORLD.RGB': meltingpot_specs.rgb(128, 256, name='WORLD.RGB'),
})
SCENARIO_OBSERVATION_SPEC = immutabledict.immutabledict({
    # Observations present in all scenarios.
    'RGB': meltingpot_specs.OBSERVATION['RGB'],
})


def _values_from_specs(
    specs: Sequence[tree.Structure[dm_env.specs.Array]]
) -> tree.Structure[np.ndarray]:
  values = tree.map_structure(lambda spec: spec.generate_value(), specs)
  return tuple(
      tree.map_structure(lambda v, n=n: np.full_like(v, n), value)
      for n, value in enumerate(values))


_AnySubstrate = TypeVar('_AnySubstrate', bound=substrate.Substrate)


def _build_mock_substrate(
    *,
    spec: Type[_AnySubstrate],
    num_players: int,
    timestep_spec: dm_env.TimeStep,
    action_spec: dm_env.specs.DiscreteArray,
) -> ...:
  """Returns a mock Substrate for use in testing.

  Args:
    spec: the Substrate class to use as a spec.
    num_players: the number of players in the substrate.
    timestep_spec: the timestep spec for a single player.
    action_spec: the action spec for a single player.
  """
  mock_substrate = mock.create_autospec(spec=spec, instance=True, spec_set=True)
  mock_substrate.__enter__.return_value = mock_substrate
  mock_substrate.__exit__.return_value = None

  mock_substrate.observation_spec.return_value = (
      timestep_spec.observation,) * num_players
  mock_substrate.reward_spec.return_value = (
      timestep_spec.reward,) * num_players
  mock_substrate.discount_spec.return_value = timestep_spec.discount
  mock_substrate.action_spec.return_value = (action_spec,) * num_players

  mock_substrate.events.return_value = ()

  observation = _values_from_specs(
      (timestep_spec.observation,) * num_players)
  mock_substrate.observation.return_value = observation
  mock_substrate.reset.return_value = dm_env.TimeStep(
      step_type=dm_env.StepType.FIRST,
      reward=(timestep_spec.reward.generate_value(),) * num_players,
      discount=0.,
      observation=observation,
  )
  mock_substrate.step.return_value = dm_env.transition(
      reward=tuple(float(i) for i in range(num_players)),
      observation=observation,
  )
  return mock_substrate


def build_mock_substrate(
    *,
    num_players: int = 8,
    num_actions: int = 8,
    observation_spec: Mapping[str,
                              dm_env.specs.Array] = SUBSTRATE_OBSERVATION_SPEC,
) -> ...:
  """Returns a mock Substrate for use in testing.

  Args:
    num_players: the number of players in the substrate.
    num_actions: the number of actions supported by the substrate.
    observation_spec: the observation spec for a single player.
  """
  return _build_mock_substrate(
      spec=substrate.Substrate,
      num_players=num_players,
      action_spec=meltingpot_specs.action(num_actions),
      timestep_spec=meltingpot_specs.timestep(observation_spec),
  )


def build_mock_substrate_like(name: str, *,
                              num_players: Optional[int] = None) -> ...:
  """Returns a mock of a specific Substrate for use in testing.

  Args:
    name: substrate to mock.
    num_players: number of players to support.
  """
  factory = meltingpot.substrate.get_factory(name)
  if num_players is None:
    num_players = len(factory.default_player_roles())
  return _build_mock_substrate(
      spec=substrate.Substrate,
      num_players=num_players,
      action_spec=factory.action_spec(),
      timestep_spec=factory.timestep_spec(),
  )


def build_mock_scenario(
    *,
    num_players: int = 8,
    num_actions: int = 8,
    observation_spec: Mapping[str,
                              dm_env.specs.Array] = SCENARIO_OBSERVATION_SPEC,
) -> ...:
  """Returns a mock Scenario for use in testing.

  Args:
    num_players: the number of focal players in the scenario.
    num_actions: the number of actions supported by the scenario.
    observation_spec: the observation spec for a single focal player.
  """
  return _build_mock_substrate(
      spec=scenario.Scenario,
      num_players=num_players,
      action_spec=meltingpot_specs.action(num_actions),
      timestep_spec=meltingpot_specs.timestep(observation_spec),
  )


def build_mock_scenario_like(name: str) -> ...:
  """Returns a mock of a specific Scenario for use in testing.

  Args:
    name: scenario to mock.
  """
  factory = meltingpot.scenario.get_factory(name)
  return _build_mock_substrate(
      spec=scenario.Scenario,
      num_players=factory.num_focal_players(),
      action_spec=factory.action_spec(),
      timestep_spec=factory.timestep_spec(),
  )



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for DMLab2D Game Objects."""

import copy
import enum
from typing import List, Mapping, NamedTuple, Optional, Sequence, Tuple, Union

import numpy as np

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes


# Type of a GameObject prefab configuration: A recursive string mapping.
# pytype: disable=not-supported-yet
PrefabConfig = Mapping[str, "PrefabConfigValue"]
PrefabConfigValue = Union[str, float, List["PrefabConfigValue"], PrefabConfig]
# pytype: enable=not-supported-yet


class Position(NamedTuple):
  x: int
  y: int


class Orientation(enum.Enum):
  NORTH = "N"
  EAST = "E"
  SOUTH = "S"
  WEST = "W"


class Transform(NamedTuple):
  position: Position
  orientation: Optional[Orientation] = None


# Special char to prefab mappings
TYPE_ALL = "all"
TYPE_CHOICE = "choice"


def get_named_components(
    game_object_config: PrefabConfig,
    name: str):
  return [component for component in game_object_config["components"]
          if component["component"] == name]


def get_first_named_component(
    game_object_config: PrefabConfig,
    name: str):
  named = get_named_components(game_object_config, name)
  if not named:
    raise ValueError(f"No component with name '{name}' found.")
  return named[0]


def build_game_objects(
    num_players: int,
    ascii_map: str,
    prefabs: Optional[Mapping[str, PrefabConfig]] = None,
    char_prefab_map: Optional[PrefabConfig] = None,
    player_palettes: Optional[Sequence[shapes.Color]] = None,
    use_badges: bool = False,
    badge_palettes: Optional[Sequence[shapes.Color]] = None,
) -> Tuple[List[PrefabConfig], List[PrefabConfig]]:
  """Build all avatar and normal game objects based on the config and map."""
  game_objects = get_game_objects_from_map(ascii_map, char_prefab_map, prefabs)
  avatar_objects = build_avatar_objects(num_players, prefabs, player_palettes)
  if use_badges:
    game_objects += build_avatar_badges(num_players, prefabs, badge_palettes)
  return game_objects, avatar_objects


def build_avatar_objects(
    num_players: int,
    prefabs: Optional[Mapping[str, PrefabConfig]] = None,
    player_palettes: Optional[Sequence[shapes.Color]] = None,
) -> List[PrefabConfig]:
  """Build all avatar and their associated game objects from the prefabs."""
  if not prefabs or "avatar" not in prefabs:
    raise ValueError(
        "Building avatar objects requested, but no avatar prefab provided.")

  if not player_palettes:
    player_palettes = [
        shapes.get_palette(colors.palette[i]) for i in range(num_players)]

  avatar_objects = []
  for idx in range(0, num_players):
    game_object = copy.deepcopy(prefabs["avatar"])
    color_palette = player_palettes[idx]
    # Lua is 1-indexed.
    lua_index = idx + 1
    # First, modify the prefab's sprite name.
    sprite_name = get_first_named_component(
        game_object, "Appearance")["kwargs"]["spriteNames"][0]
    new_sprite_name = sprite_name + str(lua_index)
    get_first_named_component(
        game_object,
        "Appearance")["kwargs"]["spriteNames"][0] = new_sprite_name
    # Second, name the same sprite in the prefab's stateManager.
    state_configs = get_first_named_component(
        game_object,
        "StateManager")["kwargs"]["stateConfigs"]
    for state_config in state_configs:
      if "sprite" in state_config and state_config["sprite"] == sprite_name:
        state_config["sprite"] = new_sprite_name
    # Third, override the prefab's color palette for this sprite.
    get_first_named_component(
        game_object, "Appearance")["kwargs"]["palettes"][0] = color_palette
    # Fourth, override the avatar's player id.
    get_first_named_component(
        game_object, "Avatar")["kwargs"]["index"] = lua_index
    avatar_objects.append(game_object)

  return avatar_objects


def build_avatar_badges(
    num_players: int,
    prefabs: Optional[Mapping[str, PrefabConfig]] = None,
    badge_palettes: Optional[Sequence[shapes.Color]] = None,
) -> List[PrefabConfig]:
  """Build all avatar and their associated game objects from the prefabs."""
  if not prefabs or "avatar_badge" not in prefabs:
    raise ValueError(
        "Building avatar badges requested, but no avatar_badge prefab " +
        "provided.")
  game_objects = []

  if badge_palettes is None:
    badge_palettes = [
        shapes.get_palette(colors.palette[i]) for i in range(num_players)]

  for idx in range(0, num_players):
    lua_index = idx + 1
    # Add the overlaid badge on top of each avatar.
    badge_object = copy.deepcopy(prefabs["avatar_badge"])
    sprite_name = get_first_named_component(
        badge_object, "Appearance")["kwargs"]["spriteNames"][0]
    new_sprite_name = sprite_name + str(lua_index)
    get_first_named_component(
        badge_object,
        "Appearance")["kwargs"]["spriteNames"][0] = new_sprite_name
    get_first_named_component(
        badge_object,
        "StateManager")["kwargs"]["stateConfigs"][0]["sprite"] = (
            new_sprite_name)
    get_first_named_component(
        badge_object, "AvatarConnector")["kwargs"]["playerIndex"] = lua_index
    get_first_named_component(
        badge_object,
        "Appearance")["kwargs"]["palettes"][0] = badge_palettes[idx]
    game_objects.append(badge_object)

  return game_objects


def get_game_object_positions_from_map(
    ascii_map: str, char: str, orientation_mode: str = "always_north"
    ) -> Sequence[Transform]:
  """Extract the occurrences of a character in the ascii map into transforms.

  For all occurrences of the given `char`, retrieves a Transform containing the
  position and orientation of the instance.

  Args:
    ascii_map: the ascii map.
    char: the character to extract transforms from the ascii map.
    orientation_mode: select a method for choosing orientations.

  Returns:
    A list of Transforms containing all the positions and orientations of all
    occurrences of the character in the map.
  """
  transforms = []
  rows = ascii_map.split("\n")
  # Assume the first line of the string consists only of '\n'. This means we
  # need to skip the first row.
  for i, row in enumerate(rows[1:]):
    indices = [i for i, c in enumerate(row) if char == c]
    for j in indices:
      if orientation_mode == "always_north":
        orientation = Orientation.NORTH
      else:
        raise ValueError("Other orientation modes are not yet implemented.")
      transform = Transform(position=Position(j, i), orientation=orientation)
      transforms.append(transform)

  return transforms


def _create_game_object(
    prefab: PrefabConfig, transform: Transform) -> PrefabConfig:
  game_object = copy.deepcopy(prefab)
  go_transform = get_first_named_component(game_object, "Transform")
  go_transform["kwargs"] = {
      "position": (transform.position.x, transform.position.y),
      "orientation": (transform.orientation.value
                      if transform.orientation is not None
                      else Orientation.NORTH.value),
    }
  return game_object


def get_game_objects_from_map(
    ascii_map: str,
    char_prefab_map: Mapping[str, str],
    prefabs: Mapping[str, PrefabConfig],
    random: np.random.RandomState = np.random.RandomState()
) -> List[PrefabConfig]:
  """Returns a list of game object configurations from the map and prefabs.

  Each prefab will have its `Transform` component overwritten to its actual
  location (and orientation, although it is all 'N' by default) in the ASCII
  map.

  Args:
    ascii_map: The map for the level. Defines which prefab to use at each
        position in the map, which is a string defining a matrix of characters.
    char_prefab_map: A dictionary mapping characters in the ascii_map to prefab
        names.
    prefabs: A collection of named prefabs that define a GameObject
        configuration.
    random: An optional random number generator.

  Returns:
    A list of game object configurations from the map and prefabs.
  """
  game_objects = []
  for char, prefab in char_prefab_map.items():
    transforms = get_game_object_positions_from_map(ascii_map, char)
    for transform in transforms:
      if hasattr(prefab, "items"):
        assert "type" in prefab
        assert "list" in prefab
        if prefab["type"] == TYPE_ALL:
          for p in prefab["list"]:
            game_objects.append(_create_game_object(prefabs[p], transform))
        elif prefab["type"] == TYPE_CHOICE:
          game_objects.append(
              _create_game_object(prefabs[random.choice(prefab["list"])],
                                  transform))
      else:  # Typical case, since named prefab.
        game_objects.append(_create_game_object(prefabs[prefab], transform))
  return game_objects



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Scenario factory."""

import collections
from collections.abc import Collection, Mapping
from typing import Callable, Optional

import immutabledict

from meltingpot import bot as mp_bot
from meltingpot import substrate as mp_substrate
from meltingpot.configs import scenarios as scenario_configs
from meltingpot.utils.scenarios import scenario
from meltingpot.utils.scenarios import scenario_factory
from meltingpot.utils.substrates import substrate as substrate_lib

SCENARIOS = frozenset(scenario_configs.SCENARIO_CONFIGS)

SubstrateTransform = Callable[[substrate_lib.Substrate],
                              substrate_lib.Substrate]


def _scenarios_by_substrate() -> Mapping[str, Collection[str]]:
  """Returns a mapping from substrates to their scenarios."""
  scenarios_by_substrate = collections.defaultdict(list)
  for name, config in scenario_configs.SCENARIO_CONFIGS.items():
    scenarios_by_substrate[config.substrate].append(name)
  return immutabledict.immutabledict({
      substrate: frozenset(scenarios)
      for substrate, scenarios in scenarios_by_substrate.items()
  })


SCENARIOS_BY_SUBSTRATE = _scenarios_by_substrate()

PERMITTED_OBSERVATIONS = frozenset({
    # The primary visual input.
    'RGB',
    # Extra observations used in some substrates.
    'HUNGER',
    'INVENTORY',
    'MY_OFFER',
    'OFFERS',
    'READY_TO_SHOOT',
    'STAMINA',
    'VOTING',
    # An extra observation that is never necessary but could perhaps help.
    'COLLECTIVE_REWARD'
})


def get_config(name: str) -> scenario_configs.ScenarioConfig:
  """Returns the config for the specified scenario."""
  return scenario_configs.SCENARIO_CONFIGS[name]


def build(
    name: str,
    *,
    substrate_transform: Optional[SubstrateTransform] = None,
) -> scenario.Scenario:
  """Builds an instance of the specified scenario.

  Args:
    name: the scenario.
    substrate_transform: optional transform to apply to underlying substrate.
      This is intended for training purposes and should not be used during
      evaluation. If applied, the observations will not be restricted to
      PERMITTED_OBSERVATIONS.

  Returns:
    The test scenario.
  """
  config = get_config(name)
  return build_from_config(config, substrate_transform=substrate_transform)


def build_from_config(
    config: scenario_configs.ScenarioConfig,
    *,
    substrate_transform: Optional[SubstrateTransform] = None,
) -> scenario.Scenario:
  """Builds a scenario from the provided config.

  Args:
    config: bot config
    substrate_transform: optional transform to apply to underlying substrate.
      This is intended for training purposes and should not be used during
      evaluation. If applied, the observations will not be restricted to
      PERMITTED_OBSERVATIONS.

  Returns:
    The test scenario.
  """
  factory = get_factory_from_config(config)
  if substrate_transform is None:
    return factory.build()
  else:
    return factory.build_transformed(substrate_transform)


def get_factory(name: str) -> scenario_factory.ScenarioFactory:
  """Returns the factory for the specified scenario."""
  config = scenario_configs.SCENARIO_CONFIGS[name]
  return get_factory_from_config(config)


def get_factory_from_config(
    config: scenario_configs.ScenarioConfig,
) -> scenario_factory.ScenarioFactory:
  """Returns a factory from the provided config."""
  substrate = mp_substrate.get_factory(config.substrate)
  bots = {
      name: mp_bot.get_factory(name)
      for name in set().union(*config.bots_by_role.values())
  }
  return scenario_factory.ScenarioFactory(
      substrate=substrate,
      roles=config.roles,
      bots=bots,
      bots_by_role=config.bots_by_role,
      is_focal=config.is_focal,
      permitted_observations=PERMITTED_OBSERVATIONS)



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Bach or Stravinsky in the Matrix.

Example video: https://youtu.be/QstXaLjiqK4

See _Running with Scissors in the Matrix_ for a general description of the
game dynamics. Here the payoff matrix represents the Bach or Stravinsky (battle
of the sexes) game. `K = 2` resources represent "Bach" and "Stravinsky" pure
strategies.

Bach or Stravinsky is an asymmetric game. Players are assigned by their slot
id to be either row players (blue) or column players (orange). Interactions are
only resolved when they are between a row player and a column player. Otherwise,
e.g. when a row player tries to interact with another row player, then nothing
happens.

Players have the default `11 x 11` (off center) observation window.
"""

from typing import Any, Dict, Mapping, Sequence, Tuple

from ml_collections import config_dict

from meltingpot.configs.substrates import the_matrix
from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import game_object_utils
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

PrefabConfig = game_object_utils.PrefabConfig

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

# The number of resources must match the (square) size of the matrix.
NUM_RESOURCES = 2

# This color is light blue.
RESOURCE1_COLOR = (123, 231, 255, 255)
RESOURCE1_HIGHLIGHT_COLOR = (157, 217, 230, 255)
RESOURCE1_COLOR_DATA = (RESOURCE1_COLOR, RESOURCE1_HIGHLIGHT_COLOR)
# This color is light orange.
RESOURCE2_COLOR = (255, 163, 123, 255)
RESOURCE2_HIGHLIGHT_COLOR = (230, 170, 157, 255)
RESOURCE2_COLOR_DATA = (RESOURCE2_COLOR, RESOURCE2_HIGHLIGHT_COLOR)

ASCII_MAP = """
WWWWWWWWWWWWWWWWWWWWWWWWW
WPPPP      W W      PPPPW
WPPPP               PPPPW
WPPPP               PPPPW
WPPPP               PPPPW
W                       W
W        11             W
W        11             W
W        aa             W
W    WW     W  222      W
WW    1a    W  222      W
WWW   1a  WWWWWWWWW     W
W     1a    111       WWW
W           111         W
W       aa W            W
W       22 W   WW       W
W       22     Waaa     W
W               222     W
W                       W
WPPPP               PPPPW
WPPPP               PPPPW
WPPPP               PPPPW
WPPPP         W     PPPPW
WWWWWWWWWWWWWWWWWWWWWWWWW
"""

_resource_names = [
    "resource_class1",
    "resource_class2",
]

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "a": {"type": "choice", "list": _resource_names},
    "1": _resource_names[0],
    "2": _resource_names[1],
    "P": "spawn_point",
    "W": "wall",
}

_COMPASS = ["N", "E", "S", "W"]

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall"],
                "spriteShapes": [shapes.WALL],
                "palettes": [{"*": (95, 95, 95, 255),
                              "&": (100, 100, 100, 255),
                              "@": (109, 109, 109, 255),
                              "#": (152, 152, 152, 255)}],
                "noRotates": [False]
            }
        },
        {
            "component": "BeamBlocker",
            "kwargs": {
                "beamType": "gameInteraction"
            }
        },
    ]
}

SPAWN_POINT = {
    "name": "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform",
        },
    ]
}

# PLAYER_COLOR_PALETTES is a list with each entry specifying the color to use
# for the player at the corresponding index.
NUM_PLAYERS_UPPER_BOUND = 32
PLAYER_COLOR_PALETTES = []
for idx in range(NUM_PLAYERS_UPPER_BOUND):
  PLAYER_COLOR_PALETTES.append(shapes.get_palette(colors.palette[idx]))

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "interact": 0}
FORWARD    = {"move": 1, "turn":  0, "interact": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "interact": 0}
BACKWARD   = {"move": 3, "turn":  0, "interact": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "interact": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "interact": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "interact": 0}
INTERACT   = {"move": 0, "turn":  0, "interact": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    INTERACT,
)

TARGET_SPRITE_SELF = {
    "name": "Self",
    "shape": shapes.CUTE_AVATAR,
    "palette": shapes.get_palette((50, 100, 200)),
    "noRotate": True,
}


def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "TheMatrix",
              "kwargs": {
                  # Prevent interaction before both interactors have collected
                  # at least one resource.
                  "disallowUnreadyInteractions": True,
                  "randomTieBreaking": True,
                  "matrix": [
                      # row player chooses a row of this matrix.
                      # B  S
                      [3, 0],  # B
                      [0, 2],  # S
                  ],
                  "columnPlayerMatrix": [
                      # column player chooses a column of this matrix.
                      # B  S
                      [2, 0],  # B
                      [0, 3],  # S
                  ],
                  "resultIndicatorColorIntervals": [
                      # red       # yellow    # green     # blue
                      (0.0, 0.5), (0.5, 1.5), (1.5, 2.5), (2.5, 3.0)
                  ],
              }
          },
          {
              "component": "StochasticIntervalEpisodeEnding",
              "kwargs": {
                  "minimumFramesPerEpisode": 1000,
                  "intervalLength": 100,  # Set equal to unroll length.
                  "probabilityTerminationPerInterval": 0.2
              }
          }
      ]
  }
  return scene


def create_resource_prefab(resource_id, color_data):
  """Creates resource prefab with provided `resource_id` (num) and color."""
  resource_name = "resource_class{}".format(resource_id)
  resource_prefab = {
      "name": resource_name,
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": resource_name,
                  "stateConfigs": [
                      {"state": resource_name + "_wait",
                       "groups": ["resourceWaits"]},
                      {"state": resource_name,
                       "layer": "lowerPhysical",
                       "sprite": resource_name + "_sprite"},
                  ]
              },
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [resource_name + "_sprite"],
                  "spriteShapes": [shapes.BUTTON],
                  "palettes": [{"*": color_data[0],
                                "#": color_data[1],
                                "x": (0, 0, 0, 0)}],
                  "noRotates": [False]
              },
          },
          {
              "component": "Resource",
              "kwargs": {
                  "resourceClass": resource_id,
                  "visibleType": resource_name,
                  "waitState": resource_name + "_wait",
                  "regenerationRate": 0.04,
                  "regenerationDelay": 10,
              },
          },
          {
              "component": "Destroyable",
              "kwargs": {
                  "waitState": resource_name + "_wait",
                  # It is possible to destroy resources but takes concerted
                  # effort to do so by zapping them `initialHealth` times.
                  "initialHealth": 3,
              },
          },
      ]
  }
  return resource_prefab


def create_prefabs() -> PrefabConfig:
  """Returns the prefabs.

  Prefabs are a dictionary mapping names to template game objects that can
  be cloned and placed in multiple locations accoring to an ascii map.
  """
  prefabs = {
      "wall": WALL,
      "spawn_point": SPAWN_POINT,
  }
  prefabs["resource_class1"] = create_resource_prefab(1, RESOURCE1_COLOR_DATA)
  prefabs["resource_class2"] = create_resource_prefab(2, RESOURCE2_COLOR_DATA)
  return prefabs


def create_avatar_object(player_idx: int,
                         color: Tuple[int, int, int],
                         row_player: bool) -> Dict[str, Any]:
  """Create an avatar object that always sees itself as blue."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  # Setup the self vs other sprite mapping.
  source_sprite_self = "Avatar" + str(lua_index)

  live_state_name = "player{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": source_sprite_self,
                       "contact": "avatar",
                       "groups": ["players"]},

                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [source_sprite_self],
                  "spriteShapes": [shapes.CUTE_AVATAR],
                  "palettes": [shapes.get_palette(color)],
                  "noRotates": [True]
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "speed": 1.0,
                  "spawnGroup": "spawnPoints",
                  "actionOrder": ["move", "turn", "interact"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "interact": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  # The following kwarg makes it possible to get rewarded even
                  # on frames when an avatar is "dead". It is needed for in the
                  # matrix games in order to correctly handle the case of two
                  # players getting hit simultaneously by the same beam.
                  "skipWaitStateRewards": False,
              }
          },
          {
              "component": "GameInteractionZapper",
              "kwargs": {
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "framesTillRespawn": 50,
                  "numResources": NUM_RESOURCES,
                  "endEpisodeOnFirstInteraction": False,
                  # Reset both players' inventories after each interaction.
                  "reset_winner_inventory": True,
                  "reset_loser_inventory": True,
                  # Both players get removed after each interaction.
                  "losingPlayerDies": True,
                  "winningPlayerDies": True,
                  # `freezeOnInteraction` is the number of frames to display the
                  # interaction result indicator, freeze, and delay delivering
                  # all results of interacting.
                  "freezeOnInteraction": 16,
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  "zapperComponent": "GameInteractionZapper",
              }
          },
          {
              "component": "InventoryObserver",
              "kwargs": {
              }
          },
          {
              "component": "DyadicRole",
              "kwargs": {
                  "rowPlayer": row_player,
              }
          },
          {
              "component": "Taste",
              "kwargs": {
                  "mostTastyResourceClass": -1,  # -1 indicates no preference.
                  # No resource is most tasty when mostTastyResourceClass == -1.
                  "mostTastyReward": 1.0,
                  "defaultTastinessReward": 0.0,
              }
          },
          {
              "component": "AvatarMetricReporter",
              "kwargs": {
                  "metrics": [
                      {
                          # Report the inventories of both players involved in
                          # an interaction on this frame formatted as
                          # (self inventory, partner inventory).
                          "name": "INTERACTION_INVENTORIES",
                          "type": "tensor.DoubleTensor",
                          "shape": (2, NUM_RESOURCES),
                          "component": "GameInteractionZapper",
                          "variable": "latest_interaction_inventories",
                      },
                      *the_matrix.get_cumulant_metric_configs(NUM_RESOURCES),
                  ]
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def create_avatar_objects(
    roles: Sequence[str],
) -> Sequence[PrefabConfig]:
  """Returns all avatar game objects."""
  avatar_objects = []
  for player_idx, role in enumerate(roles):
    if role == "default":
      if player_idx % 2 == 0:
        row_player = True
        color = (50, 100, 200)
      elif player_idx % 2 == 1:
        row_player = False
        color = (200, 100, 50)
    else:
      if role == "bach_fan":
        row_player = True
        color = (50, 100, 200)
      elif role == "stravinsky_fan":
        row_player = False
        color = (200, 100, 50)
    avatar = create_avatar_object(player_idx, color, row_player)
    avatar_objects.append(avatar)
    readiness_marker = the_matrix.create_ready_to_interact_marker(player_idx)
    avatar_objects.append(readiness_marker)
  return avatar_objects


def get_config():
  """Default configuration."""
  config = config_dict.ConfigDict()

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "INVENTORY",
      "READY_TO_SHOOT",
      # Debug only (do not use the following observations in policies).
      "INTERACTION_INVENTORIES",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "INVENTORY": specs.inventory(2),
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "INTERACTION_INVENTORIES": specs.interaction_inventories(2),
      "WORLD.RGB": specs.rgb(192, 200),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default", "bach_fan", "stravinsky_fan"})
  config.default_player_roles = ("bach_fan",) * 4 + ("stravinsky_fan",) * 4

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given roles."""
  del config
  num_players = len(roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="the_matrix",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      # Define upper bound of episode length since episodes end stochastically.
      maxEpisodeLengthFrames=5000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": create_avatar_objects(roles=roles),
          "scene": create_scene(),
          "prefabs": create_prefabs(),
          "charPrefabMap": CHAR_PREFAB_MAP,
      }
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Factory of the Commons."""

from typing import Any, Dict, Generator, Mapping, Sequence

from ml_collections import config_dict

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

_COMPASS = ["N", "E", "S", "W"]
INVISIBLE = (0, 0, 0, 0)

GRASP_SHAPE = """
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xoxxxxox
xxooooxx
"""

FLOOR_MARKING = {
    "name":
        "floor_marking",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "floor_marking",
                "stateConfigs": [{
                    "state": "floor_marking",
                    "layer": "lowestPhysical",
                    "sprite": "floor_marking",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["floor_marking"],
                "spriteShapes": [shapes.FLOOR_MARKING],
                "palettes": [shapes.DISPENSER_BELT_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

PINK_CUBE_DISPENSING_ANIMATION = {
    "name":
        "pink_cube_dispensing_animation",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                        "layer": "overlay",
                    },
                    {
                        "state": "pink_cube_dispensing_1",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_1",
                    },
                    {
                        "state": "pink_cube_dispensing_2",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_2",
                    },

                    {
                        "state": "pink_cube_dispensing_3",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_3",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["pink_cube_dispensing_1",
                                "pink_cube_dispensing_2",
                                "pink_cube_dispensing_3"],
                "spriteShapes": [shapes.CUBE_DISPENSING_ANIMATION_1,
                                 shapes.CUBE_DISPENSING_ANIMATION_2,
                                 shapes.CUBE_DISPENSING_ANIMATION_3],
                "palettes": [{
                    "a": (255, 174, 182, 255),
                    "A": (240, 161, 169, 255),
                    "&": (237, 140, 151, 255),
                    "x": (0, 0, 0, 0),
                }] * 3,
                "noRotates": [True] * 3,
            }
        },
        {
            "component": "ObjectDispensingAnimation",
            "kwargs": {
                "frameOne": "pink_cube_dispensing_1",
                "frameTwo": "pink_cube_dispensing_2",
                "frameThree": "pink_cube_dispensing_3",
                "waitState": "waitState",
            }
        },
    ]
}

DISPENSER_INDICATOR_PINK_CUBE = {
    "name":
        "dispenser_indicator_pink_cube",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "dispenser_pink_cube",
                "stateConfigs": [
                    {
                        "state": "dispenser_pink_cube",
                        "layer": "midPhysical",
                        "sprite": "dispenser_pink_cube",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["dispenser_pink_cube"],
                "spriteShapes": [shapes.HOPPER_INDICATOR_SINGLE_BLOCK],
                "palettes": [{
                    "x": (0, 0, 0, 0),
                    "a": (255, 174, 182, 255),
                }],
                "noRotates": [False]
            }
        },
        {
            "component": "DispenserIndicator",
            "kwargs": {
                "objectOne": "PinkCube",
                "objectTwo": "NoneNeeded",
            }
        }
    ]
}


SPAWN_POINT = {
    "name":
        "spawnPoint",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "spawnPoint",
                "stateConfigs": [{
                    "state": "spawnPoint",
                    "layer": "alternateLogic",
                    "groups": ["spawnPoints"]
                }],
            }
        },
        {
            "component": "Transform"
        },
    ]
}


def get_blue_cube(initial_state: str):
  """Get a blue cube prefab."""
  prefab = {
      "name":
          "blue_cube_live",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": initial_state,
                  "stateConfigs": [
                      {
                          "state": "blue_cube",
                          "layer": "lowerPhysical",
                          "sprite": "blue_cube",
                      },
                      {
                          "state": "blue_jump",
                          "layer": "lowerPhysical",
                          "sprite": "blue_jump",
                      },
                      {
                          "state": "blue_cube_drop_one",
                          "layer": "lowerPhysical",
                          "sprite": "blue_cube_drop_one",
                      },
                      {
                          "state": "blue_cube_drop_two",
                          "layer": "lowerPhysical",
                          "sprite": "blue_cube_drop_two",
                      },
                      {
                          "state": "waitState",
                      }
                  ],
              }
          },
          {
              "component": "Transform"
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["blue_cube", "blue_cube_drop_one",
                                  "blue_cube_drop_two", "blue_jump"],
                  "spriteShapes": [shapes.BLOCK,
                                   shapes.BLOCK_DROP_1,
                                   shapes.BLOCK_DROP_2,
                                   shapes.CUBE_DISPENSING_ANIMATION_1],
                  "palettes": [shapes.FACTORY_OBJECTS_PALETTE,] * 4,
                  "noRotates": [True] * 4
              }
          },
          {
              "component": "Receivable",
              "kwargs": {
                  "waitState": "waitState",
                  "liveState": "blue_cube",
              }
          },
          {
              "component": "ReceiverDropAnimation",
              "kwargs": {
                  "dropOne": "blue_cube_drop_one",
                  "dropTwo": "blue_cube_drop_two",
              }
          },
          {
              "component": "Token",
              "kwargs": {
                  "type": "BlueCube"
              }
          },
          {
              "component": "ObjectJumpAnimation",
              "kwargs": {
                  "jump": "blue_jump",
                  "drop": "blue_cube",
                  "waitState": "waitState",
              }
          },
          {
              "component": "Graspable",
              "kwargs": {
                  "graspableStates": ("blue_cube",),
                  "disconnectStates": (
                      "blue_jump", "blue_cube_drop_one", "blue_cube_drop_two",
                      "waitState",),
              }
          }
      ]
  }
  return prefab


BANANA = {
    "name":
        "banana",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "waitState",
                "stateConfigs": [
                    {
                        "state": "banana",
                        "layer": "lowerPhysical",
                        "sprite": "banana",
                    },
                    {
                        "state": "banana_jump",
                        "layer": "lowerPhysical",
                        "sprite": "banana_jump",
                    },
                    {
                        "state": "banana_drop_one",
                        "layer": "lowerPhysical",
                        "sprite": "banana_drop_one",
                    },
                    {
                        "state": "banana_drop_two",
                        "layer": "lowerPhysical",
                        "sprite": "banana_drop_two",
                    },
                    {
                        "state": "waitState"
                    }
                ],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["banana", "banana_drop_one", "banana_drop_two",
                                "banana_jump"],
                "spriteShapes": [shapes.BANANA,
                                 shapes.BANANA_DROP_1,
                                 shapes.BANANA_DROP_2,
                                 shapes.BANANA,],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE,] * 4,
                "noRotates": [True] * 4
            }
        },
        {
            "component": "Receivable",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "banana",
            }
        },
        {
            "component": "ReceiverDropAnimation",
            "kwargs": {
                "dropOne": "banana_drop_one",
                "dropTwo": "banana_drop_two",
            }
        },
        {
            "component": "Token",
            "kwargs": {
                "type": "Banana"
            }
        },
        {
            "component": "SecondObjectJumpAnimation",
            "kwargs": {
                "jump": "banana",
                "drop": "banana",
                "waitState": "waitState",
            }
        },
        {
            "component": "Graspable",
            "kwargs": {
                "graspableStates": ("banana",),
                "disconnectStates": (
                    "banana_jump", "banana_drop_one", "banana_drop_two",
                    "waitState",),
            }
        }
    ]
}

APPLE = {
    "name":
        "apples",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                    },
                    {
                        "state": "apple",
                        "layer": "appleLayer",
                        "sprite": "apple",
                    },
                    {
                        "state": "apple_jump_state",
                        "layer": "appleLayer",
                        "sprite": "apple_jump_sprite",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["apple", "apple_jump_sprite"],
                "spriteShapes": [shapes.APPLE, shapes.APPLE_JUMP],
                "palettes": [shapes.APPLE_RED_PALETTE] * 2,
                "noRotates": [True] * 2,
            }
        },
        {
            "component": "Graspable",
            "kwargs": {
                "graspableStates": ("apple",),
                "disconnectStates": ("apple_jump_state", "waitState",),
            }
        },
        {
            "component": "AppleComponent",
            "kwargs": {
                "liveState": "apple",
                "waitState": "waitState",
                "rewardForEating": 1,
            }
        },
        {
            "component": "Token",
            "kwargs": {
                "type": "Apple"
            }
        },
        {
            "component": "ObjectJumpAnimation",
            "kwargs": {
                "jump": "apple_jump_state",
                "drop": "apple",
                "waitState": "waitState",
            }
        },
        {
            "component": "SecondObjectJumpAnimation",
            "kwargs": {
                "jump": "apple",
                "drop": "apple",
                "waitState": "waitState",
            }
        },

    ]
}

PINK_CUBE = {
    "name":
        "pink_cube",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "waitState",
                "stateConfigs": [
                    {
                        "state": "pink_cube",
                        "layer": "lowerPhysical",
                        "sprite": "pink_cube",
                    },
                    {
                        "state": "pink_cube_drop_one",
                        "layer": "lowerPhysical",
                        "sprite": "pink_cube_drop_one",
                    },
                    {
                        "state": "pink_cube_drop_two",
                        "layer": "lowerPhysical",
                        "sprite": "pink_cube_drop_two",
                    },
                    {
                        "state": "pink_jump",
                        "layer": "lowerPhysical",
                        "sprite": "pink_jump",
                    },
                    {
                        "state": "waitState",
                    }
                ],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["pink_cube", "pink_cube_drop_one",
                                "pink_cube_drop_two", "pink_jump"],
                "spriteShapes": [shapes.BLOCK,
                                 shapes.BLOCK_DROP_1,
                                 shapes.BLOCK_DROP_2,
                                 shapes.CUBE_DISPENSING_ANIMATION_1],
                "palettes": [{
                    "a": (255, 174, 182, 255),
                    "A": (240, 161, 169, 255),
                    "&": (237, 140, 151, 255),
                    "x": (0, 0, 0, 0),
                }] * 4,
                "noRotates": [True] * 4
            }
        },
        {
            "component": "Receivable",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "pink_cube",
            }
        },
        {
            "component": "Token",
            "kwargs": {
                "type": "PinkCube"
            }
        },
        {
            "component": "ReceiverDropAnimation",
            "kwargs": {
                "dropOne": "pink_cube_drop_one",
                "dropTwo": "pink_cube_drop_two",
            }
        },
        {
            "component": "ObjectJumpAnimation",
            "kwargs": {
                "jump": "pink_jump",
                "drop": "pink_cube",
                "waitState": "waitState",
            }
        },
        {
            "component": "Graspable",
            "kwargs": {
                "graspableStates": ("pink_cube",),
                "disconnectStates": (
                    "pink_cube_drop_one", "pink_cube_drop_two", "pink_jump",
                    "waitState",),
            }
        }
    ]
}

APPLE_DISPENSING = {
    "name":
        "apple_dispensing",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                        "layer": "overlay",
                    },
                    {
                        "state": "apple_dispensing_1",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_1",
                    },
                    {
                        "state": "apple_dispensing_2",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_2",
                    },

                    {
                        "state": "apple_dispensing_3",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_3",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["apple_dispensing_1", "apple_dispensing_2",
                                "apple_dispensing_3"],
                "spriteShapes": [shapes.APPLE_DISPENSING_ANIMATION_1,
                                 shapes.APPLE_DISPENSING_ANIMATION_2,
                                 shapes.APPLE_DISPENSING_ANIMATION_3],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE] * 3,
                "noRotates": [True] * 3,
            }
        },
        {
            "component": "ObjectDispensingAnimation",
            "kwargs": {
                "frameOne": "apple_dispensing_1",
                "frameTwo": "apple_dispensing_2",
                "frameThree": "apple_dispensing_3",
                "waitState": "waitState",
            }
        },
    ]
}

CUBE_APPLE_DISPENSING_ANIMATION = {
    "name":
        "cube_apple_dispensing_animation",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                        "layer": "overlay",
                    },
                    {
                        "state": "apple_dispensing_1",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_1",
                    },
                    {
                        "state": "apple_dispensing_2",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_2",
                    },
                    {
                        "state": "apple_dispensing_3",
                        "layer": "overlay",
                        "sprite": "apple_dispensing_3",
                    },
                    {
                        "state": "blue_cube_dispensing_1",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_1",
                    },
                    {
                        "state": "blue_cube_dispensing_2",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_2",
                    },
                    {
                        "state": "blue_cube_dispensing_3",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_3",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["apple_dispensing_1", "apple_dispensing_2",
                                "apple_dispensing_3", "blue_cube_dispensing_1",
                                "blue_cube_dispensing_2",
                                "blue_cube_dispensing_3"],
                "spriteShapes": [shapes.APPLE_DISPENSING_ANIMATION_1,
                                 shapes.APPLE_DISPENSING_ANIMATION_2,
                                 shapes.APPLE_DISPENSING_ANIMATION_3,
                                 shapes.CUBE_DISPENSING_ANIMATION_1,
                                 shapes.CUBE_DISPENSING_ANIMATION_2,
                                 shapes.CUBE_DISPENSING_ANIMATION_3],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE] * 6,
                "noRotates": [True] * 6,
            }
        },
        {
            "component": "DoubleObjectDispensingAnimation",
            "kwargs": {
                "frameOne": "blue_cube_dispensing_1",
                "frameTwo": "blue_cube_dispensing_2",
                "frameThree": "blue_cube_dispensing_3",
                "frameFour": "apple_dispensing_1",
                "frameFive": "apple_dispensing_2",
                "frameSix": "apple_dispensing_3",
                "waitState": "waitState",
            }
        },
    ]
}

BANANA_CUBE_DISPENSING_ANIMATION = {
    "name":
        "banana_cube_dispensing_animation",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                        "layer": "overlay",
                    },
                    {
                        "state": "banana_dispensing_1",
                        "layer": "overlay",
                        "sprite": "banana_dispensing_1",
                    },
                    {
                        "state": "banana_dispensing_2",
                        "layer": "overlay",
                        "sprite": "banana_dispensing_2",
                    },
                    {
                        "state": "banana_dispensing_3",
                        "layer": "overlay",
                        "sprite": "banana_dispensing_3",
                    },
                    {
                        "state": "blue_cube_dispensing_1",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_1",
                    },
                    {
                        "state": "blue_cube_dispensing_2",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_2",
                    },

                    {
                        "state": "blue_cube_dispensing_3",
                        "layer": "overlay",
                        "sprite": "blue_cube_dispensing_3",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["banana_dispensing_1", "banana_dispensing_2",
                                "banana_dispensing_3", "blue_cube_dispensing_1",
                                "blue_cube_dispensing_2",
                                "blue_cube_dispensing_3"],
                "spriteShapes": [shapes.BANANA_DISPENSING_ANIMATION_1,
                                 shapes.BANANA,
                                 shapes.BANANA_DISPENSING_ANIMATION_3,
                                 shapes.CUBE_DISPENSING_ANIMATION_1,
                                 shapes.CUBE_DISPENSING_ANIMATION_2,
                                 shapes.CUBE_DISPENSING_ANIMATION_3],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE] * 6,
                "noRotates": [True] * 6,
            }
        },
        {
            "component": "DoubleObjectDispensingAnimation",
            "kwargs": {
                "frameOne": "blue_cube_dispensing_1",
                "frameTwo": "blue_cube_dispensing_2",
                "frameThree": "blue_cube_dispensing_3",
                "frameFour": "banana_dispensing_1",
                "frameFive": "banana_dispensing_2",
                "frameSix": "banana_dispensing_3",
                "waitState": "waitState",
            }
        },
    ]
}

PINK_CUBE_DISPENSING = {
    "name":
        "pink_cube_dispensing",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                        "layer": "overlay",
                    },
                    {
                        "state": "pink_cube_dispensing_1",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_1",
                    },
                    {
                        "state": "pink_cube_dispensing_2",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_2",
                    },

                    {
                        "state": "pink_cube_dispensing_3",
                        "layer": "overlay",
                        "sprite": "pink_cube_dispensing_3",
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["pink_cube_dispensing_1",
                                "pink_cube_dispensing_2",
                                "pink_cube_dispensing_3"],
                "spriteShapes": [shapes.CUBE_DISPENSING_ANIMATION_1,
                                 shapes.CUBE_DISPENSING_ANIMATION_2,
                                 shapes.CUBE_DISPENSING_ANIMATION_3],
                "palettes": [{
                    "a": (255, 174, 182, 255),
                    "A": (240, 161, 169, 255),
                    "&": (237, 140, 151, 255),
                    "x": (0, 0, 0, 0),
                }] * 3,
                "noRotates": [True] * 3,
            }
        },
        {
            "component": "DoubleObjectDispensingAnimation",
            "kwargs": {
                "frameOne": "pink_cube_dispensing_1",
                "frameTwo": "pink_cube_dispensing_2",
                "frameThree": "pink_cube_dispensing_3",
                "frameFour": "waitState",
                "frameFive": "waitState",
                "frameSix": "waitState",
                "waitState": "waitState",
            }
        },
    ]
}


HOPPER_MOUTH = {
    "name":
        "hopper_mouth",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "hopper_mouth_open",
                "stateConfigs": [
                    {
                        "state": "hopper_mouth_closed",
                        "layer": "lowestPhysical",
                        "sprite": "hopper_mouth_closed",
                    },
                    {
                        "state": "hopper_mouth_closing",
                        "layer": "lowestPhysical",
                        "sprite": "hopper_mouth_closing",
                    },
                    {
                        "state": "hopper_mouth_open",
                        "layer": "lowestPhysical",
                        "sprite": "hopper_mouth_open",
                    },
                    {
                        "state": "waitState"
                    }
                ],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["hopper_mouth_closed", "hopper_mouth_closing",
                                "hopper_mouth_open"],
                "spriteShapes": [shapes.HOPPER_CLOSED,
                                 shapes.HOPPER_CLOSING,
                                 shapes.HOPPER_OPEN],
                "palettes": [shapes.FACTORY_MACHINE_BODY_PALETTE] * 3,
                "noRotates": [False] * 3
            }
        },
        {
            "component": "Receiver"
        },
        {
            "component": "HopperMouth",
            "kwargs": {
                "closed": "hopper_mouth_closed",
                "opening": "hopper_mouth_closing",
                "open": "hopper_mouth_open",
            }
        },
    ]
}

HOPPER_BODY = {
    "name":
        "hopper_body",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "hopper_body",
                "stateConfigs": [
                    {
                        "state": "hopper_body",
                        "layer": "midPhysical",
                        "sprite": "hopper_body",
                    },
                    {
                        "state": "hopper_body_activated",
                        "layer": "midPhysical",
                        "sprite": "hopper_body_activated",
                    }
                ],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["hopper_body", "hopper_body_activated"],
                "spriteShapes": [shapes.HOPPER_BODY,
                                 shapes.HOPPER_BODY_ACTIVATED],
                "palettes": [{
                    "a": (140, 129, 129, 255),
                    "b": (84, 77, 77, 255),
                    "f": (92, 98, 120, 255),
                    "g": (92, 98, 120, 255),
                    "c": (92, 98, 120, 255),
                    "x": (0, 0, 0, 0),
                }] * 2,
                "noRotates": [False] * 2
            }
        },
    ]
}


HOPPER_INDICATOR = {
    "name":
        "hopper_indicator",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "hopper_indicator_two",
                "stateConfigs": [
                    {
                        "state": "waitState",
                    },
                    {
                        "state": "hopper_indicator_one",
                        "layer": "upperPhysical",
                        "sprite": "hopper_indicator_one",
                        "groups": ["indicator"]
                    },
                    {
                        "state": "hopper_indicator_two",
                        "layer": "upperPhysical",
                        "sprite": "hopper_indicator_two",
                        "groups": ["indicator"]
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames":
                    [
                        "hopper_indicator_two", "hopper_indicator_one",
                    ],
                "spriteShapes": [
                    shapes.HOPPER_INDICATOR_TWO_BLOCKS,
                    shapes.HOPPER_INDICATOR_ONE_BLOCK,],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE] * 2,
                "noRotates": [False] * 2
            }
        },
        {
            "component": "ReceiverIndicator",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "hopper_indicator_two",
                "secondLiveState": "hopper_indicator_one",
                "count": "Double",
                "type": "TwoBlocks",
            }
        }
    ]
}

HOPPER_INDICATOR_BLUE_CUBE = {
    "name":
        "hopper_indicator_blue_cube",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                    },
                    {
                        "state": "blue_cube_indicator",
                        "layer": "upperPhysical",
                        "sprite": "blue_cube_indicator",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["blue_cube_indicator"],
                "spriteShapes": [shapes.HOPPER_INDICATOR_SINGLE_BLOCK],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "ReceiverIndicator",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "blue_cube_indicator",
                "secondLiveState": "waitState",
                "count": "Single",
                "type": "BlueCube"
            }
        }
    ]
}

HOPPER_INDICATOR_BANANA = {
    "name":
        "hopper_indicator_banana",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "hopper_banana",
                "stateConfigs": [
                    {
                        "state": "hopper_banana",
                        "layer": "upperPhysical",
                        "sprite": "hopper_banana",
                    },
                    {
                        "state": "waitState"
                    }
                ]
                }
            },
        {
            "component": "ReceiverIndicator",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "hopper_banana",
                "secondLiveState": "waitState",
                "count": "Single",
                "type": "Banana",
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["hopper_banana"],
                "spriteShapes": [shapes.HOPPER_INDICATOR_SINGLE_BANANA],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

HOPPER_INDICATOR_PINK_CUBE = {
    "name":
        "hopper_indicator_pink_cube",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "waitState",
                "stateConfigs": [
                    {
                        "state": "waitState",
                    },
                    {
                        "state": "hopper_pink_cube",
                        "layer": "upperPhysical",
                        "sprite": "hopper_pink_cube",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["hopper_pink_cube"],
                "spriteShapes": [shapes.HOPPER_INDICATOR_SINGLE_BLOCK],
                "palettes":
                    [{
                        "x": (0, 0, 0, 0),
                        "a": (255, 174, 182, 255),
                    }],
                "noRotates": [False]
            }
        },
        {
            "component": "ReceiverIndicator",
            "kwargs": {
                "waitState": "waitState",
                "liveState": "hopper_pink_cube",
                "secondLiveState": "waitState",
                "count": "Single",
                "type": "PinkCube",
            }
        }
    ]
}

DISPENSER_INDICATOR_BANANA_CUBE = {
    "name":
        "dispenser_indicator_banana_cube",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "banana_cube",
                "stateConfigs": [
                    {
                        "state": "banana_cube",
                        "layer": "midPhysical",
                        "sprite": "banana_cube",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["banana_cube"],
                "spriteShapes": [shapes.HOPPER_INDICATOR_ON],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "DispenserIndicator",
            "kwargs": {
                "objectOne": "BlueCube",
                "objectTwo": "Banana",
            }
        }
    ]
}

DISPENSER_INDICATOR_CUBE_APPLE = {
    "name":
        "dispenser_indicator_cube_apple",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "cube_apple",
                "stateConfigs": [
                    {
                        "state": "cube_apple",
                        "layer": "midPhysical",
                        "sprite": "cube_apple",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["cube_apple"],
                "spriteShapes": [shapes.APPLE_CUBE_INDICATOR],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "DispenserIndicator",
            "kwargs": {
                "objectOne": "Apple",
                "objectTwo": "BlueCube",
            }
        }
    ]
}

DISPENSER_INDICATOR_APPLE = {
    "name":
        "dispenser_indicator_apple",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "dispenser_indicator_apple",
                "stateConfigs": [
                    {
                        "state": "dispenser_indicator_apple",
                        "layer": "midPhysical",
                        "sprite": "dispenser_indicator_apple",
                        "groups": ["indicator"]
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames":
                    [
                        "dispenser_indicator_apple",
                    ],
                "spriteShapes": [
                    shapes.APPLE_INDICATOR],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "DispenserIndicator",
            "kwargs": {
                "objectOne": "Apple",
                "objectTwo": "NoneNeeded",
            }
        }
    ]
}

DISPENSER_INDICATOR_TWO_APPLES = {
    "name":
        "dispenser_indicator_two_apples",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "two_apples",
                "stateConfigs": [
                    {
                        "state": "two_apples",
                        "layer": "midPhysical",
                        "sprite": "two_apples",
                    },
                ]
                }
            },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["two_apples"],
                "spriteShapes": [shapes.DOUBLE_APPLE_INDICATOR],
                "palettes": [shapes.FACTORY_OBJECTS_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "DispenserIndicator",
            "kwargs": {
                "objectOne": "Apple",
                "objectTwo": "Apple",
            }
        }
    ]
}

DISPENSER_BODY = {
    "name":
        "dispenser_body",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "dispenser_body",
                "stateConfigs": [
                    {
                        "state": "dispenser_body",
                        "layer": "lowerPhysical",
                        "sprite": "dispenser_body",
                        "groups": ["dispenser"]
                    },
                    {
                        "state": "dispenser_body_activated",
                        "layer": "lowerPhysical",
                        "sprite": "dispenser_body_activated",
                        "groups": ["dispenser"]
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": [
                    "dispenser_body",
                    "dispenser_body_activated",
                ],
                "spriteShapes": [
                    shapes.DISPENSER_BODY,
                    shapes.DISPENSER_BODY_ACTIVATED,
                ],
                "palettes": [shapes.FACTORY_MACHINE_BODY_PALETTE] * 2,
                "noRotates": [False] * 2
            }
        },
    ]
}

DISPENSER_BELT = {
    "name":
        "dispenser_belt",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "dispenser_belt_deactivated",
                "stateConfigs": [
                    {
                        "state": "dispenser_belt_deactivated",
                        "layer": "lowestPhysical",
                        "sprite": "dispenser_belt_deactivated",
                        "groups": ["dispenser"]
                    },
                    {
                        "state": "dispenser_belt_on_position_1",
                        "layer": "lowestPhysical",
                        "sprite": "dispenser_belt_on_position_1",
                        "groups": ["dispenser"]
                    },
                    {
                        "state": "dispenser_belt_on_position_2",
                        "layer": "lowestPhysical",
                        "sprite": "dispenser_belt_on_position_2",
                        "groups": ["dispenser"]
                    },
                    {
                        "state": "dispenser_belt_on_position_3",
                        "layer": "lowestPhysical",
                        "sprite": "dispenser_belt_on_position_3",
                        "groups": ["dispenser"]
                    },
                ]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": [
                    "dispenser_belt_deactivated",
                    "dispenser_belt_on_position_1",
                    "dispenser_belt_on_position_2",
                    "dispenser_belt_on_position_3",
                ],
                "spriteShapes": [
                    shapes.DISPENSER_BELT_OFF,
                    shapes.DISPENSER_BELT_ON_POSITION_1,
                    shapes.DISPENSER_BELT_ON_POSITION_2,
                    shapes.DISPENSER_BELT_ON_POSITION_3,
                ],
                "palettes": [shapes.DISPENSER_BELT_PALETTE] * 4,
                "noRotates": [False] * 4
            }
        },
        {
            "component": "ConveyerBeltOnAnimation",
            "kwargs": {
                "waitState": "dispenser_belt_deactivated",
                "stateOne": "dispenser_belt_on_position_1",
                "stateTwo": "dispenser_belt_on_position_2",
                "stateThree": "dispenser_belt_on_position_3",
            }
        }
    ]
}


NW_WALL_CORNER = {
    "name":
        "nw_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "nw_wall_corner",
                "stateConfigs": [{
                    "state": "nw_wall_corner",
                    "layer": "lowerPhysical",
                    "sprite": "NwWallCorner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NwWallCorner"],
                "spriteShapes": [shapes.NW_PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

NE_WALL_CORNER = {
    "name":
        "ne_wall_corner",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "ne_wall_corner",
                "stateConfigs": [{
                    "state": "ne_wall_corner",
                    "layer": "upperPhysical",
                    "sprite": "NeWallCorner",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NeWallCorner"],
                "spriteShapes": [shapes.NE_PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

WALL_HORIZONTAL = {
    "name":
        "wall_horizontal",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wall_horizontal",
                "stateConfigs": [{
                    "state": "wall_horizontal",
                    "layer": "lowerPhysical",
                    "sprite": "WallHorizontal",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallHorizontal"],
                "spriteShapes": [shapes.PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

WALL_T_COUPLING = {
    "name":
        "wall_t_coupling",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wall_t_coupling",
                "stateConfigs": [{
                    "state": "wall_t_coupling",
                    "layer": "upperPhysical",
                    "sprite": "WallTCoupling",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallTCoupling"],
                "spriteShapes": [shapes.PERSPECTIVE_WALL_T_COUPLING],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

WALL_EAST = {
    "name":
        "wall_east",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wall_east",
                "stateConfigs": [{
                    "state": "wall_east",
                    "layer": "lowerPhysical",
                    "sprite": "WallEast",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallEast"],
                "spriteShapes": [shapes.E_PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

WALL_WEST = {
    "name":
        "wall_west",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wall_west",
                "stateConfigs": [{
                    "state": "wall_west",
                    "layer": "lowerPhysical",
                    "sprite": "WallWest",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallWest"],
                "spriteShapes": [shapes.W_PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

WALL_MIDDLE = {
    "name":
        "wall_middle",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "wall_middle",
                "stateConfigs": [{
                    "state": "wall_middle",
                    "layer": "lowerPhysical",
                    "sprite": "WallMiddle",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["WallMiddle"],
                "spriteShapes": [shapes.MID_PERSPECTIVE_WALL],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
        {"component": "BeamBlocker", "kwargs": {"beamType": "hold"}},
        {"component": "BeamBlocker", "kwargs": {"beamType": "shove"}},
    ]
}

THRESHOLD = {
    "name":
        "threshold",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "threshold",
                "stateConfigs": [{
                    "state": "threshold",
                    "layer": "lowestPhysical",
                    "sprite": "Threshold",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Threshold"],
                "spriteShapes": [shapes.PERSPECTIVE_THRESHOLD],
                "palettes": [shapes.PERSPECTIVE_WALL_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

TILED_FLOOR = {
    "name":
        "tiled_floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "tiled_floor",
                "stateConfigs": [{
                    "state": "tiled_floor",
                    "layer": "background",
                    "sprite": "tiled_floor",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["tiled_floor"],
                "spriteShapes": [shapes.METAL_FLOOR_DOUBLE_SPACED],
                "palettes": [shapes.FACTORY_FLOOR_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

FLOOR_MARKING = {
    "name":
        "floor_marking",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "floor_marking",
                "stateConfigs": [{
                    "state": "floor_marking",
                    "layer": "lowestPhysical",
                    "sprite": "floor_marking",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["floor_marking"],
                "spriteShapes": [shapes.FLOOR_MARKING],
                "palettes": [shapes.DISPENSER_BELT_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

FLOOR_MARKING_TOP = {
    "name":
        "floor_marking_top",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "floor_marking_top",
                "stateConfigs": [{
                    "state": "floor_marking_top",
                    "layer": "lowestPhysical",
                    "sprite": "floor_marking_top",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["floor_marking_top"],
                "spriteShapes": [shapes.FLOOR_MARKING_LONG_TOP],
                "palettes": [shapes.DISPENSER_BELT_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

FLOOR_MARKING_BOTTOM = {
    "name":
        "floor_marking_bottom",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState":
                    "floor_marking_bottom",
                "stateConfigs": [{
                    "state": "floor_marking_bottom",
                    "layer": "lowestPhysical",
                    "sprite": "floor_marking_bottom",
                }],
            }
        },
        {
            "component": "Transform"
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["floor_marking_bottom"],
                "spriteShapes": [shapes.FLOOR_MARKING_LONG_BOTTOM],
                "palettes": [shapes.DISPENSER_BELT_PALETTE],
                "noRotates": [False]
            }
        },
    ]
}

human_readable_colors = list(colors.human_readable)
target_sprite_color = human_readable_colors.pop(0)
grappling_target_color_palette = shapes.get_palette(target_sprite_color)
# Add character mappings to avatar pallete for Magic Beam overlay
grappling_target_color_palette["P"] = (196, 77, 190, 130)
grappling_target_color_palette["p"] = (184, 72, 178, 80)
TARGET_SPRITE_SELF = {
    "default": {
        "name": "Self",
        "shape": shapes.CUTE_AVATAR,
        "palette": shapes.get_palette(target_sprite_color),
        "noRotate": True,
    },
    "grappling": {
        "name": "SelfGrappling",
        "shape": shapes.CUTE_AVATAR_ARMS_UP,
        "palette": grappling_target_color_palette,
        "noRotate": True,
    },
    "grappled": {
        "name": "SelfGrappled",
        "shape": shapes.MAGIC_GRAPPLED_AVATAR,
        "palette": grappling_target_color_palette,
        "noRotate": True,
    },
}

# PREFABS is a dictionary mapping names to template game objects that can
# be cloned and placed in multiple locations accoring to an ascii map.graspable
PREFABS = {
    "spawn_point": SPAWN_POINT,
    # Graspable objects.
    "apple": APPLE,
    "blue_cube_live": get_blue_cube(initial_state="blue_cube"),
    "pink_cube": PINK_CUBE,
    "blue_cube_wait": get_blue_cube(initial_state="waitState"),
    "banana": BANANA,
    # Dynamic components.
    "hopper_body": HOPPER_BODY,
    "hopper_mouth": HOPPER_MOUTH,
    # Hopper indicators.
    "hopper_indicator": HOPPER_INDICATOR,
    "hopper_indicator_pink_cube": HOPPER_INDICATOR_PINK_CUBE,
    "hopper_indicator_blue_cube": HOPPER_INDICATOR_BLUE_CUBE,
    "hopper_indicator_banana": HOPPER_INDICATOR_BANANA,
    # Dispenser indicators.
    "dispenser_indicator_apple": DISPENSER_INDICATOR_APPLE,
    "dispenser_indicator_two_apples": DISPENSER_INDICATOR_TWO_APPLES,
    "dispenser_indicator_pink_cube": DISPENSER_INDICATOR_PINK_CUBE,
    "dispenser_indicator_banana_cube": DISPENSER_INDICATOR_BANANA_CUBE,
    "dispenser_indicator_cube_apple": DISPENSER_INDICATOR_CUBE_APPLE,
    "dispenser_body": DISPENSER_BODY,
    "dispenser_belt": DISPENSER_BELT,
    "apple_dispensing_animation": APPLE_DISPENSING,
    "pink_cube_dispensing_animation": PINK_CUBE_DISPENSING_ANIMATION,
    "banana_cube_dispensing_animation": BANANA_CUBE_DISPENSING_ANIMATION,
    "cube_apple_dispensing_animation": CUBE_APPLE_DISPENSING_ANIMATION,
    # Static components.
    "nw_wall_corner": NW_WALL_CORNER,
    "ne_wall_corner": NE_WALL_CORNER,
    "wall_horizontal": WALL_HORIZONTAL,
    "wall_t_coupling": WALL_T_COUPLING,
    "wall_east": WALL_EAST,
    "wall_west": WALL_WEST,
    "wall_middle": WALL_MIDDLE,
    "threshold": THRESHOLD,
    "tiled_floor": TILED_FLOOR,
    "floor_marking": FLOOR_MARKING,
    "floor_marking_top": FLOOR_MARKING_TOP,
    "floor_marking_bottom": FLOOR_MARKING_BOTTOM,
}

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
FORWARD    = {"move": 1, "turn":  0, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
STEP_RIGHT = {"move": 2, "turn":  0, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
BACKWARD   = {"move": 3, "turn":  0, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
STEP_LEFT  = {"move": 4, "turn":  0, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
TURN_LEFT  = {"move": 0, "turn": -1, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
TURN_RIGHT = {"move": 0, "turn":  1, "pickup": 0, "grasp": 0, "hold": 0, "shove":  0}
PICKUP     = {"move": 0, "turn":  0, "pickup": 1, "grasp": 0, "hold": 0, "shove":  0}
GRASP      = {"move": 0, "turn":  0, "pickup": 0, "grasp": 1, "hold": 0, "shove":  0}
HOLD       = {"move": 0, "turn":  0, "pickup": 0, "grasp": 0, "hold": 1, "shove":  0}
# Notice that SHOVE includes both `hold` and `shove` parts.
SHOVE      = {"move": 0, "turn":  0, "pickup": 0, "grasp": 0, "hold": 1, "shove":  1}
PULL       = {"move": 0, "turn":  0, "pickup": 0, "grasp": 0, "hold": 1, "shove": -1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    PICKUP,
    GRASP,
    HOLD,
    SHOVE,
    PULL,
)


def create_scene():
  """Creates the global scene."""
  scene = {
      "name":
          "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform"
          },
          {
              "component": "StochasticIntervalEpisodeEnding",
              "kwargs": {
                  "minimumFramesPerEpisode": 1000,
                  "intervalLength": 100,  # Set equal to unroll length.
                  "probabilityTerminationPerInterval": 0.1
              }
          }
      ]
  }

  return scene


def _create_stamina_overlay(player_idx: int,
                            max_stamina_bar_states: int,
                            ) -> Generator[Dict[str, Any], None, None]:
  """Create stamina marker overlay objects."""
  # Lua is 1-indexed.
  lua_idx = player_idx + 1

  stamina_bar_state_configs = [
      # Invisible inactive (dead) overlay type.
      {"state": "staminaBarWait"},
  ]
  stamina_bar_sprite_names = []
  stamina_bar_sprite_shapes = []

  # Each player's stamina bars must be in their own layer so they do not
  # interact/collide with other players' stamina bars.
  stamina_bar_layer = f"superOverlay_{player_idx}"

  # Declare one state per level of the stamina bar.
  for i in range(max_stamina_bar_states):
    sprite_name = f"sprite_for_level_{i}"
    stamina_bar_state_configs.append(
        {"state": f"level_{i}",
         "layer": stamina_bar_layer,
         "sprite": sprite_name})
    stamina_bar_sprite_names.append(sprite_name)
    xs = "\nxxxxxxxx"
    blank_space = xs * 7
    number_of_rs = max(6 - i, 0)
    number_of_ys = i if i < 7 else 12 - i
    number_of_gs = max(i - 6, 0)
    if i >= 13:
      level = blank_space + xs
    else:
      level = (
          blank_space
          + "\nx"
          + "G" * number_of_gs
          + "Y" * number_of_ys
          + "R" * number_of_rs
          + "x"
      )
    empty = "\n".join(["x" * 8] * 8)
    # Replace the east/south/west sprites with invisible sprites so the only
    # stamina bar rendered is the one in the direction that the current player
    # is facing.
    stamina_bar_sprite_shapes.append((level, empty, empty, empty))

  # Create a stamina bar for each compass direction. Only the direction the
  # current player is facing is visible.
  for direction in ("N", "E", "S", "W"):
    yield {
        "name": "avatar_stamina_bar",
        "components": [
            {
                "component": "StateManager",
                "kwargs": {
                    "initialState": "staminaBarWait",
                    "stateConfigs": stamina_bar_state_configs
                }
            },
            {
                "component": "Transform",
            },
            {
                "component": "Appearance",
                "kwargs": {
                    "renderMode": "ascii_shape",
                    "spriteNames": stamina_bar_sprite_names,
                    "spriteShapes": stamina_bar_sprite_shapes,
                    "palettes": [{"G": (62, 137, 72, 255),
                                  "Y": (255, 216, 97, 255),
                                  "R": (162, 38, 51, 255),
                                  "x": INVISIBLE,}] * max_stamina_bar_states,
                    "noRotates": [True] * max_stamina_bar_states
                }
            },
            {
                "component": "StaminaBar",
                "kwargs": {
                    "playerIndex": lua_idx,
                    "waitState": "staminaBarWait",
                    "layer": stamina_bar_layer,
                    "direction": direction
                }
            },
        ]
    }


def create_avatar_object(player_idx: int,
                         target_sprite_self: Dict[str, Any],
                         max_stamina_bar_states: int) -> Dict[str, Any]:
  """Create an avatar object."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1
  # Setup the self vs other sprite mapping.
  avatar_sprite_name = "avatarSprite{}".format(lua_index)
  grappling_sprite = "AvatarGrappling" + str(lua_index)
  grappled_sprite = "AvatarGrappled" + str(lua_index)

  custom_sprite_map = {
      avatar_sprite_name: target_sprite_self["default"]["name"],
      grappling_sprite: target_sprite_self["grappling"]["name"],
      grappled_sprite: target_sprite_self["grappled"]["name"],
  }

  live_state_name = "player{}".format(lua_index)
  grappling_state_name = f"player{lua_index}_grappling"
  grappled_state_name = f"player{lua_index}_grappled"

  color_palette = shapes.get_palette(colors.palette[player_idx])
  # Add character mappings to avatar pallete for Magic Beam overlay
  color_palette["P"] = (196, 77, 190, 130)
  color_palette["p"] = (184, 72, 178, 80)
  spawn_group = "spawnPoints"

  avatar_object = {
      "name":
          "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState":
                      live_state_name,
                  "stateConfigs": [
                      # Initial player state.
                      {
                          "state": live_state_name,
                          "layer": "midPhysical",
                          "sprite": avatar_sprite_name,
                          "contact": "avatar",
                          "groups": ["players"]
                      },
                      {
                          "state": grappling_state_name,
                          "layer": "upperPhysical",
                          "sprite": grappling_sprite,
                          "contact": "avatar",
                          "groups": ["players"]
                      },
                      {
                          "state": grappled_state_name,
                          "layer": "upperPhysical",
                          "sprite": grappled_sprite,
                          "contact": "avatar",
                          "groups": ["players"]},
                      # Player wait type for times when they are zapped out.
                      {
                          "state": "playerWait",
                          "groups": ["playerWaits"]
                      },
                  ]
              }
          },
          {
              "component": "Transform"
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [avatar_sprite_name, grappling_sprite,
                                  grappled_sprite],
                  "spriteShapes": [shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR_ARMS_UP,
                                   shapes.MAGIC_GRAPPLED_AVATAR],
                  "palettes": [color_palette] * 3,
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "AdditionalSprites",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "customSpriteNames": [
                      target_sprite_self["default"]["name"],
                      target_sprite_self["grappling"]["name"],
                      target_sprite_self["grappled"]["name"],
                  ],
                  "customSpriteShapes": [
                      target_sprite_self["default"]["shape"],
                      target_sprite_self["grappling"]["shape"],
                      target_sprite_self["grappled"]["shape"],
                  ],
                  "customPalettes": [
                      target_sprite_self["default"]["palette"],
                      target_sprite_self["grappling"]["palette"],
                      target_sprite_self["grappled"]["palette"],
                  ],
                  "customNoRotates": [
                      target_sprite_self["default"]["noRotate"],
                      target_sprite_self["grappling"]["noRotate"],
                      target_sprite_self["grappled"]["noRotate"],
                  ],
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "additionalLiveStates": [grappled_state_name,
                                           grappling_state_name],
                  "waitState": "playerWait",
                  "spawnGroup": spawn_group,
                  "actionOrder": [
                      "move",
                      "turn",
                      "pickup",
                      "grasp",
                      # Grappling actions
                      "hold",
                      "shove",
                  ],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "pickup": {"default": 0, "min": 0, "max": 1},
                      "grasp": {"default": 0, "min": 0, "max": 1},
                      # Grappling actions
                      "hold": {"default": 0, "min": 0, "max": 1},
                      "shove": {"default": 0, "min": -1, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  "spriteMap": custom_sprite_map,
              }
          },
          {
              "component": "AvatarGrasp",
              "kwargs": {
                  "shape": GRASP_SHAPE,
                  "palette": color_palette,
                  "graspAction": "grasp",
                  # If multiple objects are at the same position then grasp them
                  # according to their layer in order `precedenceOrder`.
                  "precedenceOrder": ("appleLayer", "lowerPhysical",),
              }
          },
          {
              "component": "Grappling",
              "kwargs": {
                  "shape": shapes.MAGIC_BEAM,
                  "palette": shapes.MAGIC_BEAM_PALETTE,
                  "liveState": live_state_name,
                  "grappledState": grappled_state_name,
                  "grapplingState": grappling_state_name,
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  # In this case READY_TO_SHOOT will be 1 if hold is allowed and
                  # will be 0 if not.
                  "zapperComponent": "Grappling",
              }
          },
          {
              "component": "Stamina",
              "kwargs": {
                  "maxStamina": max_stamina_bar_states,
                  "classConfig": {
                      "name": "player",
                      "greenFreezeTime": 0,
                      "yellowFreezeTime": 2,
                      "redFreezeTime": 6,
                      # `decrementRate` = 0.5 means decrease stamina on every
                      # other costly step. `decrementRate` = 1 means decrease
                      # stamina on every costly step.
                      "decrementRate": 1.0,
                  },
                  "amountInvisible": 6,
                  "amountGreen": 6,
                  "amountYellow": 6,
                  "amountRed": 1,
                  "costlyActions": ["move",],
              }
          },
          {
              "component": "StaminaObservation",
              "kwargs": {
                  "staminaComponent": "Stamina",
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def create_avatar_objects(num_players: int,
                          max_stamina_bar_states: int = 19):
  """Returns list of avatar objects of length 'num_players'."""
  avatar_objects = []
  for player_idx in range(num_players):
    avatar_object = create_avatar_object(player_idx, TARGET_SPRITE_SELF,
                                         max_stamina_bar_states - 1)
    stamina_bar_objects = _create_stamina_overlay(player_idx,
                                                  max_stamina_bar_states)
    enter_obstacle = _create_enter_obstacle(player_idx)
    avatar_objects.append(avatar_object)
    avatar_objects.append(enter_obstacle)
    avatar_objects.extend(stamina_bar_objects)

  return avatar_objects


def _create_enter_obstacle(player_idx: int) -> Dict[str, Any]:
  # Lua is 1-indexed.
  lua_idx = player_idx + 1
  return {
      "name":
          "enter_obstacle",
      "components": [
          {
              "component": "StateManager",
              "kwargs":
                  {
                      "initialState": "obstacleWait",
                      "stateConfigs": [
                          {
                              "state": "obstacleWait"
                          },
                          {
                              "state": "obstacleLive",
                              "layer": "lowerPhysical",
                          }
                      ]
                  }
          },
          {
              "component": "Transform",
          },
          {
              "component": "AvatarConnector",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "aliveState": "obstacleLive",
                  "waitState": "obstacleWait"
              }
          },
      ]
  }


def get_config():
  """Default configuration for training on the factory2d level."""
  config = config_dict.ConfigDict()

  # Specify the number of players to particate in each episode (optional).
  config.recommended_num_players = 12

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
      "STAMINA",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  config.action_spec = specs.action(len(ACTION_SET))
  config.valid_roles = frozenset({"default"})

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given player roles."""
  num_players = len(roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="factory_of_the_commons",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=5000,  # The maximum possible number of frames.
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": config.layout.ascii_map,
          "gameObjects": create_avatar_objects(num_players),
          "scene": create_scene(),
          "prefabs": PREFABS,
          "charPrefabMap": config.layout.char_prefab_map,
      },
  )
  return substrate_definition



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Substrate factory."""

from collections.abc import Collection, Mapping, Sequence, Set
from typing import Callable

import dm_env

from meltingpot.utils.substrates import builder
from meltingpot.utils.substrates import substrate


class SubstrateFactory:
  """Factory for building specific substrates."""

  def __init__(
      self,
      *,
      lab2d_settings_builder: Callable[[Sequence[str]], builder.Settings],
      individual_observations: Collection[str],
      global_observations: Collection[str],
      action_table: Sequence[Mapping[str, int]],
      timestep_spec: dm_env.TimeStep,
      action_spec: dm_env.specs.DiscreteArray,
      valid_roles: Collection[str],
      default_player_roles: Sequence[str],
  ) -> None:
    """Initializes the factory.

    Args:
      lab2d_settings_builder: callable that takes a sequence of player roles and
        returns the lab2d settings for the substrate.
      individual_observations: names of the player-specific observations to make
        available to each player.
      global_observations: names of the dmlab2d observations to make available
        to all players.
      action_table: the possible actions. action_table[i] defines the dmlab2d
        action that will be forwarded to the wrapped dmlab2d environment for the
        discrete Melting Pot action i.
      timestep_spec: spec of timestep sent to a single player.
      action_spec: spec of action expected from a single player.
      valid_roles: player roles the substrate supports.
      default_player_roles: the default player roles vector that should be used
        for training.
    """
    self._lab2d_settings_builder = lab2d_settings_builder
    self._individual_observations = frozenset(individual_observations)
    self._global_observations = frozenset(global_observations)
    self._action_table = tuple(dict(row) for row in action_table)
    self._timestep_spec = timestep_spec
    self._action_spec = action_spec
    self._valid_roles = frozenset(valid_roles)
    self._default_player_roles = tuple(default_player_roles)

  def valid_roles(self) -> Set[str]:
    """Returns the roles the substrate supports."""
    return self._valid_roles

  def default_player_roles(self) -> Sequence[str]:
    """Returns the player roles used by scenarios."""
    return self._default_player_roles

  def timestep_spec(self) -> dm_env.TimeStep:
    """Returns spec of timestep sent to a single player."""
    return self._timestep_spec

  def action_spec(self) -> dm_env.specs.DiscreteArray:
    """Returns spec of action expected from a single player."""
    return self._action_spec

  def build(self, roles: Sequence[str]) -> substrate.Substrate:
    """Builds the substrate.

    Args:
      roles: the role each player will take.

    Returns:
      The constructed substrate.
    """
    return substrate.build_substrate(
        lab2d_settings=self._lab2d_settings_builder(roles),
        individual_observations=self._individual_observations,
        global_observations=self._global_observations,
        action_table=self._action_table)



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for King of the Hill.

Example video: https://youtu.be/VVAfeObAZzI

See _Capture the Flag_ for the description of the painting, zapping, and
movement mechanics, which also operate in this substrate.

In the King of the Hill substrate the goal is to control the hill region in
the center of the map. The hill is considered to be controlled by a team if at
least 80% of it has been colored in one team's color. The status of the hill is
indicated by indicator tiles around the map an in the center. Red indicator
tiles mean the red team is in control. Blue indicator tiles mean the blue team
is in control. Purple indicator tiles mean no team is in control.
"""

from typing import Any, Dict, Mapping, Optional, Sequence

from ml_collections import config_dict
import numpy as np

from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

_COMPASS = ["N", "E", "S", "W"]

ASCII_MAP = """
IIIIIIIIIIIIIIIIIIIIIII
IWWWWWWWWWWWWWWWWWWWWWI
IWPPP,PPPP,P,PPPP,PPPWI
IWPPP,,PP,,,,,PP,,PPPWI
IWPPP,,,,,,,,,,,,,PPPWI
IWP,,WW,,,,,,,,,WW,,PWI
IW,,,WWDWWWDWWW,WW,,,WI
IW,,,,,,uuuuuuu,D,,,,WI
IW,,,,WlGGGGGGGrW,,,,WI
IWHWWHWlGGGGGGGrWHWWHWI
IWHWWHWlGGGGGGGrWHWWHWI
IW,,,,DlGGGIGGGrD,,,,WI
IWHWWHWlGGGGGGGrWHWWHWI
IWHWWHWlGGGGGGGrWHWWHWI
IW,,,,WlGGGGGGGrW,,,,WI
IW,,,,D,ddddddd,,,,,,WI
IW,,,WW,WWWDWWWDWW,,,WI
IWQ,,WW,,,,,,,,,WW,,QWI
IWQQQ,,,,,,,,,,,,,QQQWI
IWQQQ,,QQ,,,,,QQ,,QQQWI
IWQQQ,QQQQ,Q,QQQQ,QQQWI
IWWWWWWWWWWWWWWWWWWWWWI
IIIIIIIIIIIIIIIIIIIIIII
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "P": {"type": "all", "list": ["spawn_point_red", "ground"]},
    "Q": {"type": "all", "list": ["spawn_point_blue", "ground"]},
    "W": "wall",
    "D": {"type": "choice",
          "list": ["destroyable_wall"] * 9 + ["destroyed_wall"]},
    "H": {"type": "choice",
          "list": ["destroyable_wall"] * 3 + ["destroyed_wall"]},
    "G": "hill",
    ",": "ground",
    "I": {"type": "all", "list": ["indicator", "indicator_frame"]},

    # Lines marking the edge of the hill.
    "u": {"type": "all", "list": ["ground", "line_north"]},
    "r": {"type": "all", "list": ["ground", "line_west"]},
    "d": {"type": "all", "list": ["ground", "line_south"]},
    "l": {"type": "all", "list": ["ground", "line_east"]},
}

RED_COLOR = (225, 55, 85, 255)
DARKER_RED_COLOR = (200, 35, 55, 255)
DARKEST_RED_COLOR = (160, 5, 25, 255)

BLUE_COLOR = (85, 55, 225, 255)
DARKER_BLUE_COLOR = (55, 35, 200, 255)
DARKEST_BLUE_COLOR = (25, 5, 160, 255)

PURPLE_COLOR = (107, 63, 160, 255)

LINE_NORTH = """
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
xxxxxxxx
oooooooo
"""
LINE_SOUTH = shapes.flip_vertical(LINE_NORTH)
LINE_EAST = """
xxxxxxxo
xxxxxxxo
xxxxxxxo
xxxxxxxo
xxxxxxxo
xxxxxxxo
xxxxxxxo
xxxxxxxo
"""
LINE_WEST = shapes.flip_horizontal(LINE_EAST)


def multiply_tuple(color_tuple, factor):
  alpha = color_tuple[3]
  return tuple([int(np.min([x * factor, alpha])) for x in color_tuple[0: 3]])

TEAMS_DATA = {
    "red": {"color": RED_COLOR,
            "spawn_group": "{}SpawnPoints".format("red")},
    "blue": {"color": BLUE_COLOR,
             "spawn_group": "{}SpawnPoints".format("blue")},
}

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall",],
                "spriteShapes": [shapes.WALL],
                "palettes": [{"*": (95, 95, 95, 255),
                              "&": (100, 100, 100, 255),
                              "@": (109, 109, 109, 255),
                              "#": (152, 152, 152, 255)}],
                "noRotates": [True]
            }
        },
        {
            "component": "AllBeamBlocker",
            "kwargs": {}
        },
    ]
}


def get_marking_line(orientation: str):
  """Return a line prefab to trace out the area of the hill."""
  if orientation == "N":
    shape = LINE_NORTH
  elif orientation == "E":
    shape = LINE_EAST
  elif orientation == "S":
    shape = LINE_SOUTH
  elif orientation == "W":
    shape = LINE_WEST
  else:
    raise ValueError(f"Unrecognized orientation: {orientation}")

  line_name = f"line_{orientation}"
  prefab = {
      "name": line_name,
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": line_name,
                  "stateConfigs": [{
                      "state": line_name,
                      "layer": "lowerPhysical",
                      "sprite": line_name,
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [line_name,],
                  "spriteShapes": [shape],
                  "palettes": [{"x": (0, 0, 0, 0),
                                "o": (75, 75, 75, 120)}],
                  "noRotates": [False]
              }
          },
      ]
  }
  return prefab


INDICATOR_FRAME = {
    "name": "indicator_frame",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "inert",
                "stateConfigs": [
                    {"state": "inert",
                     "layer": "superOverlay",
                     "sprite": "InertFrame"}
                ]
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["InertFrame"],
                "spriteShapes": [shapes.BUTTON],
                "palettes": [{"*": (0, 0, 0, 0),
                              "x": (55, 55, 55, 255),
                              "#": (0, 0, 0, 0)}],
                "noRotates": [True]
            }
        },
    ]
}


INDICATOR = {
    "name": "control_indicator",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "uncontrolled",
                "stateConfigs": [
                    {
                        "state": "uncontrolled",
                        "layer": "background",
                        "sprite": "UncontrolledIndicator",
                    },
                    {
                        "state": "red",
                        "layer": "background",
                        "sprite": "RedIndicator",
                    },
                    {
                        "state": "blue",
                        "layer": "background",
                        "sprite": "BlueIndicator",
                    }
                ]
            }
        },
        {"component": "Transform",},
        {
            "component": "Appearance",
            "kwargs": {
                "spriteNames": ["UncontrolledIndicator",
                                "RedIndicator",
                                "BlueIndicator"],
                "spriteRGBColors": [PURPLE_COLOR,
                                    DARKER_RED_COLOR,
                                    DARKER_BLUE_COLOR]
            }
        },
        {"component": "ControlIndicator",},
    ]
}


def create_ground_prefab(is_hill=False):
  """Return a prefab for a normal ground or a hill prefab."""
  if is_hill:
    sprite_names = ["RedHill", "BlueHill"]
    sprite_colors = [DARKER_RED_COLOR, DARKER_BLUE_COLOR]
    groups = ["grounds", "hills"]
    clean_groups = ["hill_clean",]
    red_groups = ["hill_red",]
    blue_groups = ["hill_blue",]
  else:
    sprite_names = ["RedGround", "BlueGround"]
    sprite_colors = [DARKEST_RED_COLOR, DARKEST_BLUE_COLOR]
    groups = ["grounds",]
    clean_groups = []
    red_groups = []
    blue_groups = []

  prefab = {
      "name": "ground",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "clean",
                  "stateConfigs": [
                      {
                          "state": "clean",
                          "layer": "alternateLogic",
                          "groups": groups + clean_groups,
                      },
                      {
                          "state": "red",
                          "layer": "alternateLogic",
                          "sprite": sprite_names[0],
                          "groups": groups + red_groups,
                      },
                      {
                          "state": "blue",
                          "layer": "alternateLogic",
                          "sprite": sprite_names[1],
                          "groups": groups + blue_groups,
                      },
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "spriteNames": sprite_names,
                  "spriteRGBColors": sprite_colors
              }
          },
          {
              "component": "GroundOrHill",
              "kwargs": {
                  # Must set the name to "Ground" so the color zapper component
                  # can find the location underneath the avatar to color it.
                  "name": "Ground",
                  "teamNames": ["red", "blue"],
                  "isHill": is_hill,
              }
          },
      ]
  }
  return prefab


def create_destroyable_wall_prefab(initial_state):
  """Return destroyable wall prefab, potentially starting in destroyed state."""
  if initial_state == "destroyed":
    initial_health = 0
  else:
    initial_health = 5
  prefab = {
      "name": "destroyableWall",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": initial_state,
                  "stateConfigs": [
                      {
                          "state": "destroyable",
                          "layer": "upperPhysical",
                          "sprite": "DestroyableWall",
                      },
                      {
                          "state": "damaged",
                          "layer": "upperPhysical",
                          "sprite": "DamagedWall",
                      },
                      {
                          "state": "destroyed",
                          "layer": "alternateLogic",
                          "sprite": "Rubble",
                      },
                  ],
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["DestroyableWall",
                                  "DamagedWall",
                                  "Rubble"],
                  "spriteShapes": [shapes.WALL,
                                   shapes.WALL,
                                   shapes.WALL],
                  "palettes": [{"*": (55, 55, 55, 255),
                                "&": (100, 100, 100, 255),
                                "@": (109, 109, 109, 255),
                                "#": (152, 152, 152, 255)},
                               {"*": (55, 55, 55, 255),
                                "&": (100, 100, 100, 255),
                                "@": (79, 79, 79, 255),
                                "#": (152, 152, 152, 255)},
                               {"*": (0, 0, 0, 255),
                                "&": (0, 0, 0, 255),
                                "@": (29, 29, 29, 255),
                                "#": (0, 0, 0, 255)}],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "Destroyable",
              "kwargs": {"hitNames": ["red", "blue"],
                         "initialHealth": initial_health,
                         "damagedHealthLevel": 2}
          }
      ]
  }
  return prefab


def create_spawn_point_prefab(team):
  """Return a team-specific spawn-point prefab."""
  prefab = {
      "name": "spawn_point",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "playerSpawnPoint",
                  "stateConfigs": [{
                      "state": "playerSpawnPoint",
                      "layer": "logic",
                      "groups": [TEAMS_DATA[team]["spawn_group"]],
                  }],
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "invisible",
                  "spriteNames": [],
                  "spriteRGBColors": []
              }
          },
      ]
  }
  return prefab


# PREFABS is a dictionary mapping names to template game objects that can
# be cloned and placed in multiple locations accoring to an ascii map.
PREFABS = {
    "wall": WALL,
    "spawn_point_red": create_spawn_point_prefab("red"),
    "spawn_point_blue": create_spawn_point_prefab("blue"),
    "destroyable_wall": create_destroyable_wall_prefab("destroyable"),
    "destroyed_wall": create_destroyable_wall_prefab("destroyed"),
    "hill": create_ground_prefab(is_hill=True),
    "ground": create_ground_prefab(is_hill=False),
    "indicator": INDICATOR,
    "indicator_frame": INDICATOR_FRAME,
    "line_north": get_marking_line("N"),
    "line_east": get_marking_line("E"),
    "line_south": get_marking_line("S"),
    "line_west": get_marking_line("W"),
}

# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "fireZap": 0}
FORWARD    = {"move": 1, "turn":  0, "fireZap": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "fireZap": 0}
BACKWARD   = {"move": 3, "turn":  0, "fireZap": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "fireZap": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "fireZap": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "fireZap": 0}
FIRE_ZAP_A = {"move": 0, "turn":  0, "fireZap": 1}
FIRE_ZAP_B = {"move": 0, "turn":  0, "fireZap": 2}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    FIRE_ZAP_A,  # a short-range beam with a wide area of effect
    FIRE_ZAP_B,  # a longer range beam with a thin area of effect
)


# The Scene is a non-physical object, its components implement global logic.
def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {"component": "Transform",},
          {
              "component": "HillManager",
              "kwargs": {
                  "percentToCapture": 80,
                  "rewardPerStepInControl": 1.0,
              }
          }
      ]
  }
  return scene


def create_avatar_object(
    player_idx: int,
    team: str,
    override_taste_kwargs: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
  """Create an avatar object."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  team_color = TEAMS_DATA[team]["color"]

  health1_avatar_sprite_name = "avatarSprite{}Health1".format(lua_index)
  health2_avatar_sprite_name = "avatarSprite{}Health2".format(lua_index)
  health3_avatar_sprite_name = "avatarSprite{}Health3".format(lua_index)

  health1_color_palette = shapes.get_palette(multiply_tuple(team_color, 0.35))
  health2_color_palette = shapes.get_palette(team_color)
  health3_color_palette = shapes.get_palette(multiply_tuple(team_color, 1.75))

  taste_kwargs = {
      # select `mode` from:
      #   ("none", "control_hill", "paint_hill", "zap_while_in_control")
      "mode": "none",
      "rewardAmount": 0.0,
      "zeroMainReward": False,
      "minFramesBetweenHillRewards": 0,
  }
  if override_taste_kwargs:
    taste_kwargs.update(override_taste_kwargs)

  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "health2",
                  "stateConfigs": [
                      {"state": "health1",
                       "layer": "upperPhysical",
                       "sprite": health1_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},
                      {"state": "health2",
                       "layer": "upperPhysical",
                       "sprite": health2_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},
                      {"state": "health3",
                       "layer": "upperPhysical",
                       "sprite": health3_avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},

                      # Player wait state used when they have been zapped out.
                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {"component": "Transform",},
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [health1_avatar_sprite_name,
                                  health2_avatar_sprite_name,
                                  health3_avatar_sprite_name],
                  "spriteShapes": [shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR,
                                   shapes.CUTE_AVATAR],
                  "palettes": [health1_color_palette,
                               health2_color_palette,
                               health3_color_palette],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": "health2",
                  "additionalLiveStates": ["health1", "health3"],
                  "waitState": "playerWait",
                  "spawnGroup": TEAMS_DATA[team]["spawn_group"],
                  "actionOrder": ["move",
                                  "turn",
                                  "fireZap"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "fireZap": {"default": 0, "min": 0, "max": 2},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
                  # The following kwarg makes it possible to get rewarded for
                  # team rewards even when an avatar is "dead".
                  "skipWaitStateRewards": False,
              }
          },
          {
              "component": "ColorZapper",
              "kwargs": {
                  "team": team,
                  # The color zapper beam is somewhat transparent.
                  "color": (team_color[0], team_color[1], team_color[2], 150),
                  "cooldownTime": 2,
                  "beamLength": 3,
                  "beamRadius": 1,
                  "secondaryBeamCooldownTime": 4,
                  "secondaryBeamLength": 6,
                  "secondaryBeamRadius": 0,
                  "aliveStates": ["health1", "health2", "health3"],
              }
          },
          {
              "component": "ReadyToShootObservation",
              "kwargs": {
                  "zapperComponent": "ColorZapper",
              }
          },
          {
              "component": "ZappedByColor",
              "kwargs": {
                  "team": team,
                  "allTeamNames": ["red", "blue"],
                  "framesTillRespawn": 80,
                  "penaltyForBeingZapped": 0,
                  "rewardForZapping": 0,
                  "healthRegenerationRate": 0.05,
                  "maxHealthOnGround": 2,
                  "maxHealthOnOwnColor": 3,
                  "maxHealthOnEnemyColor": 1,
                  "groundLayer": "alternateLogic",
              }
          },
          {
              "component": "TeamMember",
              "kwargs": {"team": team}
          },
          {
              "component": "Taste",
              "kwargs": taste_kwargs
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def _even_vs_odd_team_assignment(num_players,
                                 taste_kwargs: Optional[Any] = None):
  """Assign players with even ids to red team and odd ids to blue team."""
  avatar_objects = []
  for player_idx in range(0, num_players):
    if player_idx % 2 == 0:
      team = "red"
    elif player_idx % 2 == 1:
      team = "blue"
    game_object = create_avatar_object(player_idx, team,
                                       override_taste_kwargs=taste_kwargs)
    avatar_objects.append(game_object)

  return avatar_objects


def _low_vs_high_team_assignment(num_players,
                                 taste_kwargs: Optional[Any] = None):
  """Assign players with id below the median id to blue and above it to red."""
  median = np.median(range(num_players))
  avatar_objects = []
  for player_idx in range(0, num_players):
    if player_idx < median:
      team = "blue"
    elif player_idx > median:
      team = "red"
    game_object = create_avatar_object(player_idx, team,
                                       override_taste_kwargs=taste_kwargs)
    avatar_objects.append(game_object)

  return avatar_objects


def create_avatar_objects(num_players,
                          taste_kwargs: Optional[Any] = None,
                          fixed_teams: Optional[bool] = False):
  """Returns list of avatar objects of length 'num_players'."""
  assert num_players % 2 == 0, "num players must be divisible by 2"
  if fixed_teams:
    avatar_objects = _low_vs_high_team_assignment(num_players,
                                                  taste_kwargs=taste_kwargs)
  else:
    avatar_objects = _even_vs_odd_team_assignment(num_players,
                                                  taste_kwargs=taste_kwargs)
  return avatar_objects


def get_config():
  """Default configuration."""
  config = config_dict.ConfigDict()

  # If shaping_kwargs are None then use the default reward structure in which
  # all positive rewards come from your team being in control of the hill and
  # all negative rewwards come from the opposing team being in control of the
  # hill. The default reward structure is zero sum.
  config.shaping_kwargs = None

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(184, 184),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 8

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given player roles."""
  num_players = len(roles)
  substrate_definition = dict(
      levelName="paintball__king_of_the_hill",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      maxEpisodeLengthFrames=1000,
      spriteSize=8,
      topology="BOUNDED",  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": ASCII_MAP,
          "gameObjects": create_avatar_objects(
              num_players, taste_kwargs=config.shaping_kwargs),
          "scene": create_scene(),
          "prefabs": PREFABS,
          "charPrefabMap": CHAR_PREFAB_MAP,
      },
  )
  return substrate_definition



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrapper that converts action dictionary to a one hot vector."""

import functools
from typing import Mapping, Sequence, TypeVar, Union

import dm_env
import immutabledict
import numpy as np

from meltingpot.utils.substrates.wrappers import observables

T = TypeVar('T')
Numeric = Union[int, float, np.ndarray]


def _validate_action(
    action: Mapping[str, np.ndarray],
    action_spec: Mapping[str, dm_env.specs.Array]) -> None:
  """Raises ValueError if action does not matches the action_spec."""
  if set(action) != set(action_spec):
    raise ValueError('Keys do not match.')
  for key, spec in action_spec.items():
    spec.validate(action[key])


def _validate_action_table(
    action_table: Sequence[Mapping[str, np.ndarray]],
    action_spec: Mapping[str, dm_env.specs.Array]) -> None:
  """Raises ValueError if action_table does not matches the action_spec."""
  if not action_table:
    raise ValueError('action_table must not be empty')
  for action_index, action in enumerate(action_table):
    try:
      _validate_action(action, action_spec)
    except ValueError:
      raise ValueError(f'Action {action_index} ({action}) does not match '
                       f'action_spec ({action_spec}).') from None


def _immutable_action(
    action: Mapping[str, Numeric],
    action_spec: Mapping[str, dm_env.specs.Array],
) -> Mapping[str, np.ndarray]:
  """Returns an immutable action."""
  new_action = {}
  for key, value in action.items():
    if isinstance(value, np.ndarray):
      value = np.copy(value)
    else:
      value = np.array(value, dtype=action_spec[key].dtype)
    value.flags.writeable = False
    new_action[key] = value
  return immutabledict.immutabledict(new_action)


def _immutable_action_table(
    action_table: Sequence[Mapping[str, Numeric]],
    action_spec: Mapping[str, dm_env.specs.Array],
) -> Sequence[Mapping[str, np.ndarray]]:
  """Returns an immutable action table."""
  return tuple(
      _immutable_action(action, action_spec) for action in action_table)


class Wrapper(observables.ObservableLab2dWrapper):
  """Wrapper that maps a discrete action to an entry in an a table."""

  def __init__(self, env, action_table: Sequence[Mapping[str, Numeric]]):
    """Constructor.

    Args:
      env: environment to wrap. When the adaptor closes env will also be closed.
        Note that each player must have the same action spec.
      action_table: Actions that are permissable. The same action lookup is
        used by each player. action_table[i] defines the action that will be
        forwarded to the wrapped environment for discrete action i.
    """
    action_spec = env.action_spec()
    if any(action_spec[0] != spec for spec in action_spec[1:]):
      raise ValueError('Environment has heterogeneous action specs.')
    super().__init__(env)
    self._action_table = _immutable_action_table(action_table, action_spec[0])
    _validate_action_table(self._action_table, action_spec[0])

  def step(self, action: Sequence[int]):
    """See base class."""
    action = [self._action_table[player_action] for player_action in action]
    return super().step(action)

  @functools.lru_cache(maxsize=1)
  def action_spec(self) -> Sequence[dm_env.specs.DiscreteArray]:
    """See base class."""
    spec = dm_env.specs.DiscreteArray(
        num_values=len(self._action_table),
        dtype=np.int64,
        name='action')
    return tuple(spec for _ in super().action_spec())



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bot policy implementations."""

import abc
from typing import Generic, Tuple, TypeVar

import dm_env

State = TypeVar('State')


class Policy(Generic[State], metaclass=abc.ABCMeta):
  """Abstract base class for a policy.

  Must not possess any mutable state not in `initial_state`.
  """

  @abc.abstractmethod
  def initial_state(self) -> State:
    """Returns the initial state of the agent.

    Must not have any side effects.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def step(self, timestep: dm_env.TimeStep,
           prev_state: State) -> Tuple[int, State]:
    """Steps the agent.

    Must not have any side effects.

    Args:
      timestep: information from the environment
      prev_state: the previous state of the agent.

    Returns:
      action: the action to send to the environment.
      next_state: the state for the next step call.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def close(self) -> None:
    """Closes the policy."""
    raise NotImplementedError()

  def __enter__(self):
    return self

  def __exit__(self, *args, **kwargs):
    del args, kwargs
    self.close()
